#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import (
    ConcreteModel,
    Block,
    Var,
    Constraint,
    value,
    Expression,
)
from pyomo.network import Port
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom, number_total_objectives
from idaes.models.unit_models import Mixer, Separator, Product, Feed
from pyomo.util.check_units import assert_units_consistent

import watertap.property_models.NaCl_prop_pack as props
from watertap.unit_models.reverse_osmosis_0D import ReverseOsmosis0D
from watertap.unit_models.pressure_exchanger import PressureExchanger
from watertap.unit_models.pressure_changer import Pump, EnergyRecoveryDevice
from watertap.examples.flowsheets.RO_with_energy_recovery.RO_with_energy_recovery import (
    build,
    set_operating_conditions,
    initialize_system,
    solve,
    optimize_set_up,
    optimize,
    display_system,
    display_state,
    display_design,
    ERDtype,
)


solver = get_solver()

# -----------------------------------------------------------------------------
class TestROwithPX:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build(erd_type="pressure_exchanger")

        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame

        # model set up
        assert isinstance(m, ConcreteModel)
        assert isinstance(m.fs, FlowsheetBlock)
        assert isinstance(m.fs.properties, props.NaClParameterBlock)
        assert isinstance(m.fs.costing, Block)

        # unit models
        fs = m.fs
        assert isinstance(fs.feed, Feed)
        assert isinstance(fs.S1, Separator)
        assert isinstance(fs.P1, Pump)
        assert isinstance(fs.PXR, PressureExchanger)
        assert isinstance(fs.P2, Pump)
        assert isinstance(fs.M1, Mixer)
        assert isinstance(fs.RO, ReverseOsmosis0D)
        assert isinstance(fs.product, Product)
        assert isinstance(fs.disposal, Product)

        # unit model options
        # separator
        assert isinstance(fs.S1.P1, Port)
        assert isinstance(fs.S1.PXR, Port)
        # mixer
        assert isinstance(fs.M1.P1, Port)
        assert isinstance(fs.M1.P2, Port)
        assert isinstance(fs.M1.pressure_equality_constraints, Constraint)
        # RO
        assert isinstance(fs.RO.deltaP, Var)

        # additional expressions
        assert isinstance(fs.costing.annual_water_production, Expression)
        assert isinstance(fs.costing.specific_energy_consumption, Expression)

        # costing blocks
        for blk_str in ("RO",):
            blk = getattr(fs, blk_str)
            c_blk = getattr(blk, "costing")
            assert isinstance(c_blk, Block)
            assert isinstance(getattr(c_blk, "capital_cost"), Var)
            assert isinstance(getattr(c_blk, "fixed_operating_cost"), Var)

        for blk_str in (
            "P1",
            "P2",
            "PXR",
        ):
            blk = getattr(fs, blk_str)
            c_blk = getattr(blk, "costing")
            assert isinstance(c_blk, Block)
            assert isinstance(getattr(c_blk, "capital_cost"), Var)

        var_str_list = [
            "total_capital_cost",
            "total_operating_cost",
        ]
        for var_str in var_str_list:
            var = getattr(fs.costing, var_str)
            assert isinstance(var, Var)

        # arcs
        arc_dict = {
            fs.s01: (fs.feed.outlet, fs.S1.inlet),
            fs.s02: (fs.S1.P1, fs.P1.inlet),
            fs.s03: (fs.P1.outlet, fs.M1.P1),
            fs.s04: (fs.M1.outlet, fs.RO.inlet),
            fs.s05: (fs.RO.permeate, fs.product.inlet),
            fs.s06: (fs.RO.retentate, fs.PXR.high_pressure_inlet),
            fs.s07: (fs.PXR.high_pressure_outlet, fs.disposal.inlet),
            fs.s08: (fs.S1.PXR, fs.PXR.low_pressure_inlet),
            fs.s09: (fs.PXR.low_pressure_outlet, fs.P2.inlet),
            fs.s10: (fs.P2.outlet, fs.M1.P2),
        }
        for arc, port_tpl in arc_dict.items():
            assert arc.source is port_tpl[0]
            assert arc.destination is port_tpl[1]

        # units
        assert_units_consistent(fs)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame

        set_operating_conditions(
            m, water_recovery=0.5, over_pressure=0.3, solver=solver
        )

        # check fixed variables
        # feed
        assert m.fs.feed.pressure[0].is_fixed()
        assert value(m.fs.feed.pressure[0]) == 101325
        assert m.fs.feed.temperature[0].is_fixed()
        assert value(m.fs.feed.temperature[0]) == 298.15
        assert m.fs.feed.flow_mass_phase_comp[0, "Liq", "H2O"].is_fixed()
        assert value(m.fs.feed.flow_mass_phase_comp[0, "Liq", "H2O"]) == pytest.approx(
            0.9857, rel=1e-3
        )
        assert m.fs.feed.flow_mass_phase_comp[0, "Liq", "NaCl"].is_fixed()
        assert value(m.fs.feed.flow_mass_phase_comp[0, "Liq", "NaCl"]) == pytest.approx(
            3.575e-2, rel=1e-3
        )
        # pumps and pressure exchangers
        assert m.fs.P1.efficiency_pump[0].is_fixed()
        assert value(m.fs.P1.efficiency_pump[0]) == 0.8
        assert m.fs.P1.control_volume.properties_out[0].pressure.is_fixed()
        assert value(
            m.fs.P1.control_volume.properties_out[0].pressure
        ) == pytest.approx(7.493e6, rel=1e-3)
        assert m.fs.P2.efficiency_pump[0].is_fixed()
        assert value(m.fs.P2.efficiency_pump[0]) == 0.8
        assert m.fs.PXR.efficiency_pressure_exchanger[0].is_fixed()
        assert value(m.fs.PXR.efficiency_pressure_exchanger[0]) == 0.95
        # RO
        assert m.fs.RO.A_comp[0, "H2O"].is_fixed()
        assert value(m.fs.RO.A_comp[0, "H2O"]) == 4.2e-12
        assert m.fs.RO.B_comp[0, "NaCl"].is_fixed()
        assert value(m.fs.RO.B_comp[0, "NaCl"]) == 3.5e-8
        assert m.fs.RO.feed_side.channel_height.is_fixed()
        assert value(m.fs.RO.feed_side.channel_height) == 1e-3
        assert m.fs.RO.feed_side.spacer_porosity.is_fixed()
        assert value(m.fs.RO.feed_side.spacer_porosity) == 0.97
        assert m.fs.RO.permeate.pressure[0].is_fixed()
        assert value(m.fs.RO.permeate.pressure[0]) == 101325
        assert m.fs.RO.width.is_fixed()
        assert value(m.fs.RO.width) == 5
        assert not m.fs.RO.area.is_fixed()
        assert value(m.fs.RO.area) == pytest.approx(50, rel=1e-3)

        # check degrees of freedom
        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    def test_initialize_system(self, system_frame):
        m = system_frame

        initialize_system(m, solver=solver)

        # check results across pressure exchanger, proxy for both upstream and downstream of RO
        # high pressure inlet
        assert value(
            m.fs.PXR.high_pressure_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        ) == pytest.approx(0.4928, rel=1e-3)
        assert value(
            m.fs.PXR.high_pressure_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
        ) == pytest.approx(3.561e-2, rel=1e-3)
        assert value(m.fs.PXR.high_pressure_inlet.temperature[0]) == pytest.approx(
            298.15, rel=1e-3
        )
        assert value(m.fs.PXR.high_pressure_inlet.pressure[0]) == pytest.approx(
            7.394e6, rel=1e-3
        )
        # low pressure inlet
        assert value(
            m.fs.PXR.low_pressure_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        ) == pytest.approx(0.4980, rel=1e-3)
        assert value(
            m.fs.PXR.low_pressure_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
        ) == pytest.approx(1.806e-2, rel=1e-3)
        assert value(m.fs.PXR.low_pressure_inlet.temperature[0]) == pytest.approx(
            298.15, rel=1e-3
        )
        assert value(m.fs.PXR.low_pressure_inlet.pressure[0]) == pytest.approx(
            101325, rel=1e-3
        )
        # low pressure outlet
        assert value(m.fs.PXR.low_pressure_outlet.pressure[0]) == pytest.approx(
            7.030e6, rel=1e-3
        )

    @pytest.mark.component
    def test_simulation(self, system_frame):
        m = system_frame

        solve(m, solver=solver)

        # check system metrics
        assert value(m.fs.RO.recovery_vol_phase[0, "Liq"]) == pytest.approx(
            0.4954, rel=1e-3
        )
        assert value(m.fs.costing.specific_energy_consumption) == pytest.approx(
            2.727, rel=1e-3
        )
        assert value(m.fs.costing.LCOW) == pytest.approx(0.4394, rel=1e-3)

        # check mass balance
        assert pytest.approx(
            value(m.fs.feed.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]), rel=1e-3
        ) == value(m.fs.product.inlet.flow_mass_phase_comp[0, "Liq", "H2O"]) + value(
            m.fs.disposal.inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        )
        assert pytest.approx(
            value(m.fs.feed.outlet.flow_mass_phase_comp[0, "Liq", "NaCl"]), rel=1e-3
        ) == value(m.fs.product.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"]) + value(
            m.fs.disposal.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
        )

    @pytest.mark.component
    def test_display_system(self, system_frame, capsys):
        m = system_frame
        display_system(m)

        captured = capsys.readouterr()

        assert (
            captured.out
            == """---system metrics---
Feed: 1.02 kg/s, 35000 ppm
Product: 0.493 kg/s, 280 ppm
Volumetric recovery: 49.5%
Water recovery: 50.0%
Energy Consumption: 2.7 kWh/m3
Levelized cost of water: 0.44 $/m3
"""
        )

    @pytest.mark.component
    def test_display_design(self, system_frame, capsys):
        m = system_frame
        display_design(m)

        captured = capsys.readouterr()

        assert (
            captured.out
            == """---decision variables---
Operating pressure 74.9 bar
Membrane area 60.2 m2
---design variables---
Pump 1
outlet pressure: 74.9 bar
power 4.57 kW
Separator
Split fraction 50.53
Pump 2
outlet pressure: 74.9 bar
power 0.30 kW
"""
        )

    @pytest.mark.component
    def test_display_state(self, system_frame, capsys):
        m = system_frame
        display_state(m)

        captured = capsys.readouterr()

        assert (
            captured.out
            == """---state---
Feed      : 1.021 kg/s, 35000 ppm, 1.0 bar
Split 1   : 0.505 kg/s, 35000 ppm, 1.0 bar
P1 out    : 0.505 kg/s, 35000 ppm, 74.9 bar
Split 2   : 0.516 kg/s, 35000 ppm, 1.0 bar
PXR LP out: 0.516 kg/s, 35000 ppm, 70.3 bar
P2 out    : 0.516 kg/s, 35000 ppm, 74.9 bar
Mix out   : 1.021 kg/s, 35000 ppm, 74.9 bar
RO perm   : 0.493 kg/s, 280 ppm, 1.0 bar
RO reten  : 0.528 kg/s, 67389 ppm, 73.9 bar
PXR HP out: 0.528 kg/s, 67389 ppm, 1.0 bar
"""
        )

    @pytest.mark.component
    def test_optimization(self, system_frame):
        m = system_frame

        optimize_set_up(m)
        assert number_total_objectives(m) == 1
        optimize(m, solver=solver)

        # check decision variables
        assert value(m.fs.RO.inlet.pressure[0]) == pytest.approx(5.708e6, rel=1e-3)
        assert value(m.fs.RO.area) == pytest.approx(115, rel=1e-3)
        # check system metrics
        assert value(m.fs.RO.recovery_vol_phase[0, "Liq"]) == pytest.approx(
            0.4954, rel=1e-3
        )
        assert value(m.fs.costing.specific_energy_consumption) == pytest.approx(
            2.110, rel=1e-3
        )
        assert value(m.fs.costing.LCOW) == pytest.approx(0.4111, rel=1e-3)


class TestROwithTurbine:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build(erd_type=ERDtype.pump_as_turbine)

        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame
        fs = m.fs
        assert isinstance(fs.ERD, EnergyRecoveryDevice)

        # arcs
        arc_dict = {
            fs.s01: (fs.feed.outlet, fs.P1.inlet),
            fs.s02: (fs.P1.outlet, fs.RO.inlet),
            fs.s03: (fs.RO.permeate, fs.product.inlet),
            fs.s04: (fs.RO.retentate, fs.ERD.inlet),
            fs.s05: (fs.ERD.outlet, fs.disposal.inlet),
        }
        for arc, port_tpl in arc_dict.items():
            assert arc.source is port_tpl[0]
            assert arc.destination is port_tpl[1]

    @pytest.mark.component
    def test_units(self, system_frame):
        assert_units_consistent(system_frame)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        assert m.fs.ERD.efficiency_pump[0].is_fixed()
        assert pytest.approx(0.95, rel=1e-5) == value(m.fs.ERD.efficiency_pump[0])
        assert pytest.approx(101325, rel=1e-5) == value(
            m.fs.ERD.control_volume.properties_out[0].pressure
        )

        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    def test_initialize_system(self, system_frame):
        m = system_frame
        initialize_system(m, solver=solver)
        assert pytest.approx(60.1545, rel=1e-5) == value(m.fs.RO.area)

    @pytest.mark.component
    def test_optimize_setup(self, system_frame):
        m = system_frame
        optimize_set_up(m)
        solve(m, solver=solver)
        assert degrees_of_freedom(m) == 1
        assert number_total_objectives(m) == 1

    @pytest.mark.component
    def test_solution(self, system_frame):
        m = system_frame
        fs = m.fs
        assert pytest.approx(120.154, rel=1e-5) == value(fs.RO.area)
        assert pytest.approx(2.42916, rel=1e-5) == value(
            fs.costing.specific_energy_consumption
        )
        assert pytest.approx(1.15385, rel=1e-3) == value(
            fs.costing.specific_electrical_carbon_intensity
        )
        assert pytest.approx(0.54814, rel=1e-5) == value(fs.costing.LCOW)

    @pytest.mark.component
    def test_config_error(self, system_frame):
        with pytest.raises(Exception):
            build(erd_type="not_a_configuration")



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Error classes and utilities for the electrolyte database (EDB).
"""
__author__ = "Dan Gunter"

from typing import Dict
from pprint import pformat


class Error(Exception):
    """Abstract base class for all EDB errors."""


class ConfigGeneratorError(Error):
    """Base class of errors for ConfigGenerator actions and effects."""


class DataWrapperError(Error):
    """ "Base class of errors for DataWrapper actions and effects."""


class BadConfiguration(DataWrapperError):
    """Bad configuration provided to build a DataWrapper class."""

    def __init__(self, whoami: str, config: Dict, missing: str = None, why: str = None):
        if missing:
            why = f"Missing field '{missing}'"
        elif why is None:
            why = "Unknown error"
        dumped = pformat(config)
        msg = f"Bad configuration provided to '{whoami}': {why}.\n{dumped}"
        super().__init__(msg)


class ValidationError(Error):
    """Validation error."""

    def __init__(self, err):
        msg = f"{err}"
        super().__init__(msg)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest

from pyomo.environ import ConcreteModel, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.core.util.scaling as iscale

from watertap.unit_models.mvc.components import Condenser
import watertap.property_models.water_prop_pack as props

solver = get_solver()

# -----------------------------------------------------------------------------
@pytest.mark.component
def test_complete_condense():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.WaterParameterBlock()
    m.fs.unit = Condenser(property_package=m.fs.properties)

    # scaling
    m.fs.properties.set_default_scaling("flow_mass_phase_comp", 1, index=("Vap", "H2O"))
    m.fs.properties.set_default_scaling("flow_mass_phase_comp", 1, index=("Liq", "H2O"))
    iscale.set_scaling_factor(m.fs.unit.control_volume.heat, 1e-6)
    iscale.calculate_scaling_factors(m)

    # state variables
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Vap", "H2O"].fix(1)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(1e-8)
    m.fs.unit.inlet.temperature[0].fix(400)  # K
    m.fs.unit.inlet.pressure[0].fix(0.5e5)  # Pa

    m.fs.unit.outlet.temperature[0].fix(340)  # K

    # solving
    assert_units_consistent(m)
    assert degrees_of_freedom(m) == 0

    m.fs.unit.initialize_build()

    solver = get_solver()
    results = solver.solve(m, tee=False)
    assert_optimal_termination(results)

    # Check values
    distillate_blk = m.fs.unit.control_volume.properties_out[0]
    assert distillate_blk.flow_mass_phase_comp["Liq", "H2O"].value == pytest.approx(
        1.0000, rel=1e-3
    )
    assert distillate_blk.temperature.value == pytest.approx(340, rel=1e-3)
    assert distillate_blk.pressure.value == pytest.approx(50000, rel=1e-3)
    assert m.fs.unit.control_volume.heat[0].value == pytest.approx(-2.4358e6, rel=1e-3)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest

from watertap.edb import ElectrolyteDB


@pytest.fixture
def edb_client_server_info(edb: ElectrolyteDB) -> dict:
    """
    Perform an operation that ensures that the `edb` fixture has a client that:
    - Is able to connect to the server (non-mock), or
    - Is a "convincing fake" (mock), so that test functions using `edb` can expect a realistic behavior

    Additionally, if this fixture is dispatched before "real" users (here this is done by using a `test__` prefix with two underscores),
    it avoids any first-use inconsistencies, such as e.g. the time spent to wait for a connection
    being counted as part of the duration of the first test for which the `edb` fixture is instantiated.
    """
    return edb._client.server_info()


def test_edb_client_server_connection(edb_client_server_info: dict):
    "This should fail unless the EDB instance's client is either successfully connected or a realistic mock (i.e. not None, a dict, ...)."
    assert edb_client_server_info["ok"]



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Thermophysical property package to be used in conjunction with ADM1 reactions.
"""

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    LiquidPhase,
    Component,
    Solute,
    Solvent,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import fix_state_vars, revert_state_vars
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale

# Some more information about this module
__author__ = "Alejandro Garciadiego, Adam Atia, Xinhong Liu"
# Using Andrew Lee's formulation of ASM1 as a template

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ADM1ParameterBlock")
class ADM1ParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class
    """

    def build(self):
        """
        Callable method for Block construction.
        """
        super().build()

        self._state_block_class = ADM1StateBlock

        # Add Phase objects
        self.Liq = LiquidPhase()

        # Add Component objects
        self.H2O = Solvent()

        # All soluble components on kg COD/m^3 basis
        self.S_su = Solute(doc="Monosaccharides")
        self.S_aa = Solute(doc="Amino acids")
        self.S_fa = Solute(doc="Long chain fatty acids")
        self.S_va = Solute(doc="Total valerate")
        self.S_bu = Solute(doc="Total butyrate")
        self.S_pro = Solute(doc="Total propionate")
        self.S_ac = Solute(doc="Total acetate")
        self.S_h2 = Solute(doc="Hydrogen gas")
        self.S_ch4 = Solute(doc="Methane gas")
        self.S_IC = Solute(doc="Inorganic carbon")
        self.S_IN = Solute(doc="Inorganic nitrogen")
        self.S_I = Solute(doc="Soluble inerts")

        self.X_c = Solute(doc="Composites")
        self.X_ch = Solute(doc="Carbohydrates")
        self.X_pr = Solute(doc="Proteins")
        self.X_li = Solute(doc="Lipids")
        self.X_su = Solute(doc="Sugar degraders")
        self.X_aa = Solute(doc="Amino acid degraders")
        self.X_fa = Solute(doc="Long chain fatty acid (LCFA) degraders")
        self.X_c4 = Solute(doc="Valerate and butyrate degraders")
        self.X_pro = Solute(doc="Propionate degraders")
        self.X_ac = Solute(doc="Acetate degraders")
        self.X_h2 = Solute(doc="Hydrogen degraders")
        self.X_I = Solute(doc="Particulate inerts")

        # TODO: Additional components not in Table but referred to in text

        self.S_cat = Component(doc="Total cation equivalents concentration")
        self.S_an = Component(doc="Total anion equivalents concentration")

        # Heat capacity of water
        self.cp_mass = pyo.Param(
            initialize=4182,
            doc="Specific heat capacity of water",
            units=pyo.units.J / pyo.units.kg / pyo.units.K,
        )
        # Density of water
        self.dens_mass = pyo.Param(
            initialize=997,
            doc="Density of water",
            units=pyo.units.kg / pyo.units.m**3,
        )

        # Thermodynamic reference state
        self.pressure_ref = pyo.Param(
            within=pyo.PositiveReals,
            mutable=True,
            default=101325.0,
            doc="Reference pressure",
            units=pyo.units.Pa,
        )
        self.temperature_ref = pyo.Param(
            within=pyo.PositiveReals,
            mutable=True,
            default=298.15,
            doc="Reference temperature",
            units=pyo.units.K,
        )

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "flow_vol": {"method": None},
                "pressure": {"method": None},
                "temperature": {"method": None},
                "conc_mass_comp": {"method": None},
            }
        )
        obj.define_custom_properties(
            {
                "anions": {"method": None},
                "cations": {"method": None},
            }
        )
        obj.add_default_units(
            {
                "time": pyo.units.s,
                "length": pyo.units.m,
                "mass": pyo.units.kg,
                "amount": pyo.units.kmol,
                "temperature": pyo.units.K,
            }
        )


class _ADM1StateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(
        self,
        state_args=None,
        state_vars_fixed=False,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for property package.

        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:
            flow_mol_comp : value at which to initialize component flows (default=None)
            pressure : value at which to initialize pressure (default=None)
            temperature : value at which to initialize temperature (default=None)
            outlvl : sets output level of initialization routine
            state_vars_fixed: Flag to denote if state vars have already been fixed.
                              True - states have already been fixed and
                              initialization does not need to worry
                              about fixing and unfixing variables.
                              False - states have not been fixed. The state
                              block will deal with fixing/unfixing.
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         True - states variables are not unfixed, and
                         a dict of returned containing flags for
                         which states were fixed during initialization.
                         False - state variables are unfixed after
                         initialization by calling the release_state method.

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")

        if state_vars_fixed is False:
            # Fix state variables if not already fixed
            flags = fix_state_vars(self, state_args)

        else:
            # Check when the state vars are fixed already result in dof 0
            for k in self.keys():
                if degrees_of_freedom(self[k]) != 0:
                    raise Exception(
                        "State vars fixed but degrees of freedom "
                        "for state block is not zero during "
                        "initialization."
                    )

        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                self.release_state(flags)

        init_log.info("Initialization Complete.")

    def release_state(self, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of logging
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")

        if flags is None:
            return
        # Unfix state variables
        revert_state_vars(self, flags)
        init_log.info("State Released.")


@declare_process_block_class("ADM1StateBlock", block_class=_ADM1StateBlock)
class ADM1StateBlockData(StateBlockData):
    """
    StateBlock for calculating thermophysical properties associated with the ADM1
    reaction system.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super().build()

        # Create state variables
        self.flow_vol = pyo.Var(
            initialize=1,
            domain=pyo.NonNegativeReals,
            doc="Total volumetric flowrate",
            units=pyo.units.m**3 / pyo.units.s,
        )
        self.pressure = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=101325.0,
            bounds=(1e1, 1e6),
            doc="Pressure",
            units=pyo.units.Pa,
        )
        self.temperature = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=298.15,
            bounds=(273.15, 323.15),
            doc="Temperature",
            units=pyo.units.K,
        )
        Comp_dict = {
            "S_su": 0.012,
            "S_aa": 0.0053,
            "S_fa": 0.099,
            "S_va": 0.012,
            "S_bu": 0.013,
            "S_pro": 0.016,
            "S_ac": 0.20,
            "S_h2": 2.3e-7,
            "S_ch4": 0.055,
            "S_IC": 0.15 * 12,
            "S_IN": 0.13 * 14,
            "S_I": 0.33,
            "X_c": 0.31,
            "X_ch": 0.028,
            "X_pr": 0.10,
            "X_li": 0.029,
            "X_su": 0.42,
            "X_aa": 1.18,
            "X_fa": 0.24,
            "X_c4": 0.43,
            "X_pro": 0.14,
            "X_ac": 0.76,
            "X_h2": 0.32,
            "X_I": 25.6,
        }
        self.conc_mass_comp = pyo.Var(
            self.params.solute_set,
            domain=pyo.NonNegativeReals,
            initialize=Comp_dict,
            doc="Component mass concentrations",
            units=pyo.units.kg / pyo.units.m**3,
        )
        self.anions = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=0.02,
            doc="Anions in molar concentration",
            units=pyo.units.kmol / pyo.units.m**3,
        )
        self.cations = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=0.04,
            doc="Cations in molar concentration",
            units=pyo.units.kmol / pyo.units.m**3,
        )

        def material_flow_expression(self, j):
            if j == "H2O":
                return self.flow_vol * self.params.dens_mass
            elif j == "S_an":
                # Convert moles of anions to mass assuming all is CL-
                return (
                    self.flow_vol * self.anions * (35 * pyo.units.kg / pyo.units.kmol)
                )
            elif j == "S_cat":
                # Convert moles of cations to mass assuming all is Na+
                return (
                    self.flow_vol * self.cations * (23 * pyo.units.kg / pyo.units.kmol)
                )
            else:
                return self.flow_vol * self.conc_mass_comp[j]

        self.material_flow_expression = pyo.Expression(
            self.component_list,
            rule=material_flow_expression,
            doc="Material flow terms",
        )

        def enthalpy_flow_expression(self):
            return (
                self.flow_vol
                * self.params.dens_mass
                * self.params.cp_mass
                * (self.temperature - self.params.temperature_ref)
            )

        self.enthalpy_flow_expression = pyo.Expression(
            rule=enthalpy_flow_expression, doc="Enthalpy flow term"
        )

        def material_density_expression(self, j):
            if j == "H2O":
                return self.params.dens_mass
            elif j == "S_cat":
                # Convert moles of alkalinity to mass of catioons assuming all is Na
                return self.cations * (23 * pyo.units.kg / pyo.units.kmol)
            elif j == "S_an":
                # Convert moles of alkalinity to mass of aniona assuming all is Cl
                return self.anions * (35 * pyo.units.kg / pyo.units.kmol)
            else:
                return self.conc_mass_comp[j]

        self.material_density_expression = pyo.Expression(
            self.component_list,
            rule=material_density_expression,
            doc="Material density terms",
        )

        def energy_density_expression(self):
            return (
                self.params.dens_mass
                * self.params.cp_mass
                * (self.temperature - self.params.temperature_ref)
            )

        self.energy_density_expression = pyo.Expression(
            rule=energy_density_expression, doc="Energy density term"
        )

        iscale.set_scaling_factor(self.flow_vol, 1e5)
        iscale.set_scaling_factor(self.temperature, 1e-1)
        iscale.set_scaling_factor(self.pressure, 1e-6)
        iscale.set_scaling_factor(self.conc_mass_comp, 1e2)
        iscale.set_scaling_factor(self.conc_mass_comp["S_h2"], 1e5)
        iscale.set_scaling_factor(self.anions, 1e2)
        iscale.set_scaling_factor(self.cations, 1e2)

    def get_material_flow_terms(self, p, j):
        return self.material_flow_expression[j]

    def get_enthalpy_flow_terms(self, p):
        return self.enthalpy_flow_expression

    def get_material_density_terms(self, p, j):
        return self.material_density_expression[j]

    def get_energy_density_terms(self, p):
        return self.energy_density_expression

    def default_material_balance_type(self):
        return MaterialBalanceType.componentPhase

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def define_state_vars(self):
        return {
            "flow_vol": self.flow_vol,
            "anions": self.anions,
            "cations": self.cations,
            "conc_mass_comp": self.conc_mass_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }

    def define_display_vars(self):
        return {
            "Volumetric Flowrate": self.flow_vol,
            "Molar anions": self.anions,
            "Molar cations": self.cations,
            "Mass Concentration": self.conc_mass_comp,
            "Temperature": self.temperature,
            "Pressure": self.pressure,
        }

    def get_material_flow_basis(self):
        return MaterialFlowBasis.mass

    def calculate_scaling_factors(self):
        # Get default scale factors and do calculations from base classes
        super().calculate_scaling_factors()

        # No constraints in this model as yet, just need to set scaling factors
        # for expressions
        sf_F = iscale.get_scaling_factor(self.flow_vol, default=1, warning=True)
        sf_T = iscale.get_scaling_factor(self.temperature, default=1, warning=True)

        # Mass flow and density terms
        for j in self.component_list:
            if j == "H2O":
                sf_C = pyo.value(1 / self.params.dens_mass)
            elif j == "S_cat":
                sf_C = 1e-1 * iscale.get_scaling_factor(
                    self.cations, default=1, warning=True
                )
            elif j == "S_an":
                sf_C = 1e-1 * iscale.get_scaling_factor(
                    self.anions, default=1, warning=True
                )
            else:
                sf_C = iscale.get_scaling_factor(
                    self.conc_mass_comp[j], default=1e2, warning=True
                )

            iscale.set_scaling_factor(self.material_flow_expression[j], sf_F * sf_C)
            iscale.set_scaling_factor(self.material_density_expression[j], sf_C)

        # Enthalpy and energy terms
        sf_rho_cp = pyo.value(1 / (self.params.dens_mass * self.params.cp_mass))
        iscale.set_scaling_factor(
            self.enthalpy_flow_expression, sf_F * sf_rho_cp * sf_T
        )
        iscale.set_scaling_factor(self.energy_density_expression, sf_rho_cp * sf_T)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Param,
    Suffix,
    NonNegativeReals,
    units as pyunits,
)
from idaes.models.unit_models.separator import SeparatorData, SplittingType

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
)

from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.misc import add_object_reference
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from watertap.costing.unit_models.electroNP import cost_electroNP

__author__ = "Chenyu Wang"

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ElectroNPZO")
class ElectroNPZOdata(SeparatorData):
    """
    Zero order electrochemical nutrient removal (ElectroNP) model based on specified removal efficiencies for nitrogen and phosphorus.
    """

    CONFIG = SeparatorData.CONFIG()
    CONFIG.outlet_list = ["treated", "byproduct"]
    CONFIG.split_basis = SplittingType.componentFlow

    def build(self):
        # Call UnitModel.build to set up dynamics
        super(ElectroNPZOdata, self).build()

        if len(self.config.property_package.solvent_set) > 1:
            raise ConfigurationError(
                "ElectroNP model only supports one solvent component,"
                "the provided property package has specified {} solvent components".format(
                    len(self.config.property_package.solvent_set)
                )
            )

        if len(self.config.property_package.solvent_set) == 0:
            raise ConfigurationError(
                "The ElectroNP model was expecting a solvent and did not receive it."
            )

        if (
            len(self.config.property_package.solute_set) == 0
            and len(self.config.property_package.ion_set) == 0
        ):
            raise ConfigurationError(
                "The ElectroNP model was expecting at least one solute or ion and did not receive any."
            )

        if "treated" and "byproduct" not in self.config.outlet_list:
            raise ConfigurationError(
                "{} encountered unrecognised "
                "outlet_list. This should not "
                "occur - please use treated "
                "and byproduct.".format(self.name)
            )

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        units_meta = self.config.property_package.get_metadata().get_derived_units

        add_object_reference(self, "properties_in", self.mixed_state)
        add_object_reference(self, "properties_treated", self.treated_state)
        add_object_reference(self, "properties_byproduct", self.byproduct_state)

        # Add performance variables
        self.recovery_frac_mass_H2O = Var(
            self.flowsheet().time,
            initialize=0.99999,
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            bounds=(0.0, 1.0000001),
            doc="Mass recovery fraction of water in the treated stream",
        )
        self.recovery_frac_mass_H2O.fix()

        # Default solute concentration
        self.P_removal = Param(
            within=NonNegativeReals,
            mutable=True,
            default=0.98,
            doc="Reference phosphorus removal fraction on a mass basis",
            units=pyunits.dimensionless,
        )

        self.N_removal = Param(
            within=NonNegativeReals,
            mutable=True,
            default=0.3,
            doc="Reference ammonia removal fraction on a mass basis",
            units=pyunits.dimensionless,
        )

        add_object_reference(self, "removal_frac_mass_comp", self.split_fraction)

        @self.Constraint(
            self.flowsheet().time,
            self.config.property_package.component_list,
            doc="soluble fraction",
        )
        def split_components(blk, t, i):
            if i == "H2O":
                return (
                    blk.removal_frac_mass_comp[t, "byproduct", i]
                    == 1 - blk.recovery_frac_mass_H2O[t]
                )
            elif i == "S_PO4":
                return blk.removal_frac_mass_comp[t, "byproduct", i] == blk.P_removal
            elif i == "S_NH4":
                return blk.removal_frac_mass_comp[t, "byproduct", i] == blk.N_removal
            else:
                return blk.removal_frac_mass_comp[t, "byproduct", i] == 1e-7

        self.electricity = Var(
            self.flowsheet().time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Electricity consumption of unit",
        )

        self.energy_electric_flow_mass = Var(
            units=pyunits.kWh / pyunits.kg,
            doc="Electricity intensity with respect to phosphorus removal",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for electricity consumption based on phosphorus removal",
        )
        def electricity_consumption(b, t):
            return b.electricity[t] == (
                b.energy_electric_flow_mass
                * pyunits.convert(
                    b.properties_byproduct[t].get_material_flow_terms("Liq", "S_PO4"),
                    to_units=pyunits.kg / pyunits.hour,
                )
            )

        self.magnesium_chloride_dosage = Var(
            units=pyunits.dimensionless,
            bounds=(0, None),
            doc="Dosage of magnesium chloride per phosphorus removal",
        )

        self.MgCl2_flowrate = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            bounds=(0, None),
            doc="Magnesium chloride flowrate",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for magnesium chloride demand based on phosphorus removal.",
        )
        def MgCl2_demand(b, t):
            return b.MgCl2_flowrate[t] == (
                b.magnesium_chloride_dosage
                * pyunits.convert(
                    b.properties_byproduct[t].get_material_flow_terms("Liq", "S_PO4"),
                    to_units=pyunits.kg / pyunits.hour,
                )
            )

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Water Recovery"] = self.recovery_frac_mass_H2O[time_point]
        for j in self.config.property_package.solute_set:
            var_dict[f"Solute Removal {j}"] = self.removal_frac_mass_comp[
                time_point, "byproduct", j
            ]
        var_dict["Electricity Demand"] = self.electricity[time_point]
        var_dict["Electricity Intensity"] = self.energy_electric_flow_mass
        var_dict[
            "Dosage of magnesium chloride per treated phosphorus"
        ] = self.magnesium_chloride_dosage
        var_dict["Magnesium Chloride Demand"] = self.MgCl2_flowrate[time_point]
        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Inlet": self.inlet,
                "Treated": self.treated,
                "Byproduct": self.byproduct,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        iscale.set_scaling_factor(self.recovery_frac_mass_H2O, 1)

        if iscale.get_scaling_factor(self.energy_electric_flow_mass) is None:
            sf = iscale.get_scaling_factor(
                self.energy_electric_flow_mass, default=1e2, warning=True
            )
            iscale.set_scaling_factor(self.energy_electric_flow_mass, sf)

        if iscale.get_scaling_factor(self.magnesium_chloride_dosage) is None:
            sf = iscale.get_scaling_factor(
                self.magnesium_chloride_dosage, default=1e1, warning=True
            )
            iscale.set_scaling_factor(self.magnesium_chloride_dosage, sf)

        for (t, i, j), v in self.removal_frac_mass_comp.items():
            if i == "treated":
                for i in self.config.outlet_list:
                    if j == "S_PO4":
                        sf = 1
                    elif j == "S_NH4":
                        sf = 1
                    else:
                        sf = 1
            iscale.set_scaling_factor(v, sf)

        for (t, i, j), v in self.removal_frac_mass_comp.items():
            if i == "byproduct":
                for i in self.config.outlet_list:
                    if j == "S_PO4":
                        sf = 1
                    elif j == "S_NH4":
                        sf = 1
                    else:
                        sf = 1e7
            iscale.set_scaling_factor(v, sf)

        for t, v in self.electricity.items():
            sf = (
                iscale.get_scaling_factor(self.energy_electric_flow_mass)
                * iscale.get_scaling_factor(self.byproduct.flow_vol[t])
                * iscale.get_scaling_factor(self.byproduct.conc_mass_comp[t, "S_PO4"])
            )
            iscale.set_scaling_factor(v, sf)

        for t, v in self.MgCl2_flowrate.items():
            sf = (
                iscale.get_scaling_factor(self.magnesium_chloride_dosage)
                * iscale.get_scaling_factor(self.byproduct.flow_vol[t])
                * iscale.get_scaling_factor(self.byproduct.conc_mass_comp[t, "S_PO4"])
            )
            iscale.set_scaling_factor(v, sf)

    @property
    def default_costing_method(self):
        return cost_electroNP



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    NonNegativeReals,
    NegativeReals,
    value,
)

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    useDefault,
)
from idaes.core.util import scaling as iscale
import idaes.logger as idaeslog

from watertap.core import (  # noqa # pylint: disable=unused-import
    ConcentrationPolarizationType,
    MassTransferCoefficient,
    MembraneChannel1DBlock,
    PressureChangeType,
)
from watertap.core.membrane_channel1d import CONFIG_Template
from watertap.unit_models.reverse_osmosis_base import (
    ReverseOsmosisBaseData,
    _add_has_full_reporting,
)

__author__ = "Adam Atia, Bernard Knueven"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ReverseOsmosis1D")
class ReverseOsmosis1DData(ReverseOsmosisBaseData):
    """Standard 1D Reverse Osmosis Unit Model Class."""

    CONFIG = CONFIG_Template()

    _add_has_full_reporting(CONFIG)

    def _process_config(self):
        if self.config.transformation_method is useDefault:
            _log.warning(
                "Discretization method was "
                "not specified for the "
                "reverse osmosis module. "
                "Defaulting to finite "
                "difference method."
            )
            self.config.transformation_method = "dae.finite_difference"

        if self.config.transformation_scheme is useDefault:
            _log.warning(
                "Discretization scheme was "
                "not specified for the "
                "reverse osmosis module."
                "Defaulting to backward finite "
                "difference."
            )
            self.config.transformation_scheme = "BACKWARD"

    def _add_feed_side_membrane_channel_and_geometry(self):
        # Check configuration errors
        self._process_config()

        # Build 1D Membrane Channel
        self.feed_side = MembraneChannel1DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            area_definition=self.config.area_definition,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            transformation_method=self.config.transformation_method,
            transformation_scheme=self.config.transformation_scheme,
            finite_elements=self.config.finite_elements,
            collocation_points=self.config.collocation_points,
        )
        self._add_length_and_width()
        self.feed_side.add_geometry(length_var=self.length, width_var=self.width)
        self._add_area(include_constraint=True)

    def _add_deltaP(self):
        units_meta = self.config.property_package.get_metadata().get_derived_units
        self.deltaP = Var(
            self.flowsheet().config.time,
            initialize=-1e5,
            bounds=(-1e6, 0),
            domain=NegativeReals,
            units=units_meta("pressure"),
            doc="Pressure drop across unit",
        )
        if self.config.pressure_change_type == PressureChangeType.fixed_per_stage:

            @self.Constraint(
                self.flowsheet().config.time,
                self.length_domain,
                doc="Fixed pressure drop across unit",
            )
            def eq_pressure_drop(b, t, x):
                return b.deltaP[t] == b.length * b.feed_side.dP_dx[t, x]

        else:

            @self.Constraint(
                self.flowsheet().config.time, doc="Pressure drop across unit"
            )
            def eq_pressure_drop(b, t):
                return b.deltaP[t] == sum(
                    b.feed_side.dP_dx[t, x] * b.length / b.nfe
                    for x in b.difference_elements
                )

    def _add_mass_transfer(self):
        units_meta = self.config.property_package.get_metadata().get_derived_units

        def mass_transfer_phase_comp_initialize(b, t, x, p, j):
            return value(
                self.feed_side.properties[t, x].get_material_flow_terms("Liq", j)
                * self.recovery_mass_phase_comp[t, "Liq", j]
            )

        self.mass_transfer_phase_comp = Var(
            self.flowsheet().config.time,
            self.difference_elements,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=mass_transfer_phase_comp_initialize,
            bounds=(0.0, 1e6),
            domain=NonNegativeReals,
            units=units_meta("mass")
            * units_meta("time") ** -1
            * units_meta("length") ** -1,
            doc="Mass transfer to permeate",
        )

        # ==========================================================================
        # Mass transfer term equation
        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass transfer term",
        )
        def eq_mass_transfer_term(b, t, x, p, j):
            return (
                b.mass_transfer_phase_comp[t, x, p, j]
                == -b.feed_side.mass_transfer_term[t, x, p, j]
            )

        # ==========================================================================
        # Feed and permeate-side mass transfer connection --> Mp,j = Mf,transfer = Jj * W * L/n
        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass transfer from feed to permeate",
        )
        def eq_connect_mass_transfer(b, t, x, p, j):
            return (
                b.permeate_side[t, x].get_material_flow_terms(p, j)
                == -b.feed_side.mass_transfer_term[t, x, p, j] * b.length / b.nfe
            )

        # ==========================================================================
        # Mass flux = feed mass transfer equation
        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass transfer term",
        )
        def eq_mass_flux_equal_mass_transfer(b, t, x, p, j):
            return (
                b.flux_mass_phase_comp[t, x, p, j] * b.width
                == -b.feed_side.mass_transfer_term[t, x, p, j]
            )

        # ==========================================================================
        # Final permeate mass flow rate (of solvent and solute) --> Mp,j, final = sum(Mp,j)

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Permeate mass flow rates exiting unit",
        )
        def eq_permeate_production(b, t, p, j):
            return b.mixed_permeate[t].get_material_flow_terms(p, j) == sum(
                b.permeate_side[t, x].get_material_flow_terms(p, j)
                for x in b.difference_elements
            )

    def calculate_scaling_factors(self):
        if iscale.get_scaling_factor(self.dens_solvent) is None:
            sf = iscale.get_scaling_factor(
                self.feed_side.properties[0, 0].dens_mass_phase["Liq"]
            )
            iscale.set_scaling_factor(self.dens_solvent, sf)

        super().calculate_scaling_factors()

        # these variables should have user input, if not there will be a warning
        if iscale.get_scaling_factor(self.width) is None:
            sf = iscale.get_scaling_factor(self.width, default=1, warning=True)
            iscale.set_scaling_factor(self.width, sf)

        if iscale.get_scaling_factor(self.length) is None:
            sf = iscale.get_scaling_factor(self.length, default=10, warning=True)
            iscale.set_scaling_factor(self.length, sf)

        for (t, x, p, j), v in self.mass_transfer_phase_comp.items():
            sf = (
                iscale.get_scaling_factor(
                    self.feed_side.properties[t, x].get_material_flow_terms(p, j)
                )
                / iscale.get_scaling_factor(self.feed_side.length)
            ) * value(self.nfe)
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)
            v = self.feed_side.mass_transfer_term[t, x, p, j]
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)

        if hasattr(self, "deltaP"):
            for v in self.deltaP.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-4)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from idaes.core.util.scaling import get_scaling_factor, __none_left_mult
from idaes.core.solvers import get_solver


def assert_no_initialization_perturbation(blk, optarg=None, solver=None):
    """
    Assert that IPOPT will *not* move the initialization

    Args:
        blk: Pyomo block
        optarg: IPOPT options (default=None) (Should be None if solver is specified)
        solver: Pyomo IPOPT solver instance (default=None) (Should be None if optarg is specified).

    Returns:
        None
    """
    if solver is not None and optarg is not None:
        raise ValueError("Supply a solver or optarg, not both")
    if optarg is None:
        optarg = {}
    if solver is None:
        solver = get_solver(options=optarg)
    if solver.name not in (
        "ipopt",
        "ipopt-watertap",
    ):
        raise ValueError(f"Solver {solver.name} is not supported")

    options = solver.options
    bound_push = options.get("bound_push", 1e-2)
    bound_frac = options.get("bound_frac", 1e-2)
    bound_relax_factor = options.get("bound_relax_factor", 1e-8)
    if solver.name == "ipopt-watertap":
        user_scaling = True
    else:
        user_scaling = (
            options.get("nlp_scaling_method", "gradient-based") == "user-scaling"
        )

    for v, val, result in generate_initialization_perturbation(
        blk, bound_push, bound_frac, bound_relax_factor, user_scaling
    ):
        raise ValueError(
            f"IPOPT will move scaled initial value for variable {v.name} from {val:e} to {result:e}"
        )


def print_initialization_perturbation(
    blk, bound_push=1e-2, bound_frac=1e-2, bound_relax_factor=1e-8, user_scaling=False
):
    """
    Print the initialization perturbations performed by IPOPT for a given Block

    Args:
        blk: Pyomo block
        bound_push: bound_push to evaluate (same as IPOPT option) (default=1e-2)
        bound_frac: bound_frac to evaluate (same as IPOPT option) (default=1e-2)
        bound_relax_factor: bound_relax_factor to evaluate (same as IPOPT option) (default=1e-8)
        user_scaling: If True, the variables are scaled as if `nlp_scaling_method = user-scaling`
                       is used. (default=False)

    Returns:
        None
    """
    for v, val, result in generate_initialization_perturbation(
        blk, bound_push, bound_frac, bound_relax_factor, user_scaling
    ):
        print(
            f"IPOPT will move scaled initial value for variable {v.name} from {val:e} to {result:e}"
        )


def generate_initialization_perturbation(
    blk, bound_push=1e-2, bound_frac=1e-2, bound_relax_factor=1e-8, user_scaling=False
):
    """
    Generate the initialization perturbations performed by IPOPT for a given Block

    Args:
        blk: Pyomo block
        bound_push: bound_push to evaluate (same as IPOPT option) (default=1e-2)
        bound_frac: bound_frac to evaluate (same as IPOPT option) (default=1e-2)
        bound_relax_factor: bound_relax_factor to evaluate (same as IPOPT option) (default=1e-8)
        user_scaling: If True, the variables are scaled as if `nlp_scaling_method = user-scaling`
                       is used. (default=False)

    Yields:
        tuple: (pyo.Var object, current_value, perturbed_value)
    """
    kappa1 = bound_push
    kappa2 = bound_frac
    for v in blk.component_data_objects(pyo.Var):
        if v.value is None:
            _log.warning(f"Variable {v.name} has no initial value")
            continue
        if v.fixed:
            continue
        if user_scaling:
            sf = get_scaling_factor(v, default=1.0)
        else:
            sf = 1.0
        v_lb = __none_left_mult(v.lb, sf)
        if v_lb is not None:
            v_lb -= bound_relax_factor * max(1, abs(v_lb))
        v_value = v.value * sf
        v_ub = __none_left_mult(v.ub, sf)
        if v_ub is not None:
            v_ub += bound_relax_factor * max(1, abs(v_ub))
        if v_lb is not None:
            if v_ub is not None:
                pl = min(kappa1 * max(1, abs(v_lb)), kappa2 * (v_ub - v_lb))
            else:
                pl = kappa1 * max(1, abs(v_lb))
            if v_value < v_lb + pl:
                yield (v, v_value, v_lb + pl)
        if v_ub is not None:
            if v_lb is not None:
                pu = min(kappa1 * max(1, abs(v_ub)), kappa2 * (v_ub - v_lb))
            else:
                pu = kappa1 * max(1, abs(v_ub))
            if v_value > v_ub - pu:
                yield (v, v_value, v_ub - pu)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    check_optimal_termination,
    Param,
    Suffix,
    NonNegativeReals,
    PositiveIntegers,
    value,
    units as pyunits,
)

from pyomo.common.config import Bool, ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    MomentumBalanceType,
    EnergyBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.constants import Constants
from idaes.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError, InitializationError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from watertap.core import ControlVolume0DBlock, InitializationMixin

__author__ = "Austin Ladshaw"

_log = idaeslog.getLogger(__name__)

# Name of the unit model
@declare_process_block_class("CoagulationFlocculation")
class CoagulationFlocculationData(InitializationMixin, UnitModelBlockData):
    """
    Zero order Coagulation-Flocculation model based on Jar Tests
    """

    # CONFIG are options for the unit model
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag; must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default**: False. The filtration unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )

    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag; must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default**: False. The filtration unit does not have defined volume, thus
    this must be False.""",
        ),
    )

    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )

    CONFIG.declare(
        "is_isothermal",
        ConfigValue(
            default=True,
            domain=Bool,
            description="""Assume isothermal conditions for control volume(s); energy_balance_type must be EnergyBalanceType.none,
    **default** - True.""",
        ),
    )

    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.none,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.none.
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )

    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )

    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )

    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    CONFIG.declare(
        "chemical_additives",
        ConfigValue(
            default={},
            domain=dict,
            description="Specification of chemical additives used in coagulation process",
            doc="""
        A dict of chemical additives used in coagulation process
        along with their molecular weights, the moles of salt produced per mole of
        chemical added, and the molecular weights of the salt produced by the chemical
        additive with the format of::

            {'chem_name_1':
                {'parameter_data':
                    {
                    'mw_additive': (value, units),
                    'moles_salt_per_mole_additive': value,
                    'mw_salt': (value, units)
                    }
                },
            'chem_name_2':
                {'parameter_data':
                    {
                    'mw_additive': (value, units),
                    'moles_salt_per_mole_additive': value,
                    'mw_salt': (value, units)
                    }
                },
            }

        """,
        ),
    )

    def _validate_config(self):
        if (
            self.config.is_isothermal
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            raise ConfigurationError(
                "If the isothermal assumption is used then the energy balance type must be none"
            )

    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # this creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Next, get the base units of measurement from the property definition
        units_meta = self.config.property_package.get_metadata().get_derived_units

        # Check configs for errors
        self._validate_config()

        # check the optional config arg 'chemical_additives'
        common_msg = (
            "The 'chemical_additives' dict MUST contain a dict of 'parameter_data' for "
            + "each chemical name. That 'parameter_data' dict MUST contain 'mw_chem', "
            + "'moles_salt_per_mole_additive', and 'mw_salt' as keys. Users are also "
            + "required to provide the values for the molecular weights and the units "
            + "within a tuple arg. Example format provided below.\n\n"
            + "{'chem_name_1': \n"
            + "     {'parameter_data': \n"
            + "        {'mw_additive': (value, units), \n"
            + "         'moles_salt_per_mole_additive': value, \n"
            + "         'mw_salt': (value, units)} \n"
            + "     }, \n"
            + "}\n\n"
        )

        # Populate temp dicts for parameter and variable setting
        mw_adds = {}
        mw_salts = {}
        molar_rat = {}

        for j in self.config.chemical_additives:
            if type(self.config.chemical_additives[j]) != dict:
                raise ConfigurationError(
                    "\n Did not provide a 'dict' for chemical \n" + common_msg
                )
            if "parameter_data" not in self.config.chemical_additives[j]:
                raise ConfigurationError(
                    "\n Did not provide a 'parameter_data' for chemical \n" + common_msg
                )
            if "mw_additive" not in self.config.chemical_additives[j]["parameter_data"]:
                raise ConfigurationError(
                    "\n Did not provide a 'mw_additive' for chemical \n" + common_msg
                )
            if (
                "moles_salt_per_mole_additive"
                not in self.config.chemical_additives[j]["parameter_data"]
            ):
                raise ConfigurationError(
                    "\n Did not provide a 'moles_salt_per_mole_additive' for chemical \n"
                    + common_msg
                )
            if "mw_salt" not in self.config.chemical_additives[j]["parameter_data"]:
                raise ConfigurationError(
                    "\n Did not provide a 'mw_salt' for chemical \n" + common_msg
                )
            if (
                type(self.config.chemical_additives[j]["parameter_data"]["mw_additive"])
                != tuple
            ):
                raise ConfigurationError(
                    "\n Did not provide a tuple for 'mw_additive' \n" + common_msg
                )
            if (
                type(self.config.chemical_additives[j]["parameter_data"]["mw_salt"])
                != tuple
            ):
                raise ConfigurationError(
                    "\n Did not provide a tuple for 'mw_salt' \n" + common_msg
                )
            if not isinstance(
                self.config.chemical_additives[j]["parameter_data"][
                    "moles_salt_per_mole_additive"
                ],
                (int, float),
            ):
                raise ConfigurationError(
                    "\n Did not provide a number for 'moles_salt_per_mole_additive' \n"
                    + common_msg
                )

            mw_adds[j] = pyunits.convert_value(
                self.config.chemical_additives[j]["parameter_data"]["mw_additive"][0],
                from_units=self.config.chemical_additives[j]["parameter_data"][
                    "mw_additive"
                ][1],
                to_units=pyunits.kg / pyunits.mol,
            )
            mw_salts[j] = pyunits.convert_value(
                self.config.chemical_additives[j]["parameter_data"]["mw_salt"][0],
                from_units=self.config.chemical_additives[j]["parameter_data"][
                    "mw_salt"
                ][1],
                to_units=pyunits.kg / pyunits.mol,
            )
            molar_rat[j] = self.config.chemical_additives[j]["parameter_data"][
                "moles_salt_per_mole_additive"
            ]

        # Add unit variables
        # Linear relationship between TSS (mg/L) and Turbidity (NTU)
        #           TSS (mg/L) = Turbidity (NTU) * slope + intercept
        #   Default values come from the following paper:
        #       H. Rugner, M. Schwientek,B. Beckingham, B. Kuch, P. Grathwohl,
        #       Environ. Earth Sci. 69 (2013) 373-380. DOI: 10.1007/s12665-013-2307-1
        self.slope = Var(
            self.flowsheet().config.time,
            initialize=1.86,
            bounds=(0.0, 10),
            domain=NonNegativeReals,
            units=pyunits.mg / pyunits.L,
            doc="Slope relation between TSS (mg/L) and Turbidity (NTU)",
        )

        self.intercept = Var(
            self.flowsheet().config.time,
            initialize=0,
            bounds=(0, 10),
            domain=NonNegativeReals,
            units=pyunits.mg / pyunits.L,
            doc="Intercept relation between TSS (mg/L) and Turbidity (NTU)",
        )

        self.initial_turbidity_ntu = Var(
            self.flowsheet().config.time,
            initialize=50,
            bounds=(0, 10000),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Initial measured Turbidity (NTU) from Jar Test",
        )

        self.final_turbidity_ntu = Var(
            self.flowsheet().config.time,
            initialize=1,
            bounds=(0, 10000),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Final measured Turbidity (NTU) from Jar Test",
        )

        self.chemical_doses = Var(
            self.flowsheet().config.time,
            self.config.chemical_additives.keys(),
            initialize=0,
            bounds=(0, 100),
            domain=NonNegativeReals,
            units=pyunits.mg / pyunits.L,
            doc="Dosages of the set of chemical additives",
        )

        self.chemical_mw = Param(
            self.config.chemical_additives.keys(),
            mutable=True,
            initialize=mw_adds,
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.mol,
            doc="Molecular weights of the set of chemical additives",
        )

        self.salt_mw = Param(
            self.config.chemical_additives.keys(),
            mutable=True,
            initialize=mw_salts,
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.mol,
            doc="Molecular weights of the produced salts from chemical additives",
        )

        self.salt_from_additive_mole_ratio = Param(
            self.config.chemical_additives.keys(),
            mutable=True,
            initialize=molar_rat,
            domain=NonNegativeReals,
            units=pyunits.mol / pyunits.mol,
            doc="Moles of the produced salts from 1 mole of chemical additives",
        )

        # Below set of Vars are for the power usage of the unit
        #       User's will need to provide scaling factors for these
        # -----------------------------------------------------------
        #   Mines, R.O., Environmental engineering: Principles
        #       and Practice, 1st Ed, John Wiley & Sons, 2014.
        #       Ch. 6.
        self.rapid_mixing_retention_time = Var(
            self.flowsheet().config.time,
            initialize=30,
            bounds=(0.1, 10000),
            domain=NonNegativeReals,
            units=pyunits.s,
            doc="Hydraulic retention time of each rapid mixing basin in seconds",
        )

        self.num_rapid_mixing_basins = Var(
            initialize=1,
            bounds=(1, 10),
            domain=PositiveIntegers,
            units=pyunits.dimensionless,
            doc="Number of rapid mixing basins in series",
        )

        self.rapid_mixing_vel_grad = Var(
            self.flowsheet().config.time,
            initialize=250,
            bounds=(0.1, 10000),
            domain=NonNegativeReals,
            units=pyunits.s**-1,
            doc="Velocity gradient in each rapid mixing basin in (m/s)/m",
        )

        # NOTE: There are 2 modes for flocculation mixing discussed in literature
        #       Here we are only intially defining the 'Paddle-Wheel' mode. Other
        #       modes can be added later (if needed). The 'Paddle-Wheel' configuration
        #       is the most common used for conventional water treatment.
        self.floc_retention_time = Var(
            self.flowsheet().config.time,
            initialize=1800,
            bounds=(10, 10000),
            domain=NonNegativeReals,
            units=pyunits.s,
            doc="Hydraulic retention time of the flocculation mixing basin in seconds",
        )

        self.single_paddle_length = Var(
            initialize=2,
            bounds=(0.1, 100),
            domain=NonNegativeReals,
            units=pyunits.m,
            doc="Length of a single paddle blade (from center of rotation to the edge) in meters",
        )

        self.single_paddle_width = Var(
            initialize=0.5,
            bounds=(0.01, 100),
            domain=NonNegativeReals,
            units=pyunits.m,
            doc="Width of a single paddle blade in meters",
        )

        self.paddle_rotational_speed = Var(
            self.flowsheet().config.time,
            initialize=100,
            bounds=(0.01, 10000),
            domain=NonNegativeReals,
            units=pyunits.s**-1,
            doc="Rotational speed of the paddles in revolutions per second",
        )

        self.paddle_drag_coef = Var(
            self.flowsheet().config.time,
            initialize=1.5,
            bounds=(0.1, 10),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Drag coefficient for the paddles in the flocculation basin",
        )

        self.vel_fraction = Var(
            initialize=0.7,
            bounds=(0.6, 0.9),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Fraction of actual paddle velocity relative to local water velocity",
        )

        self.num_paddle_wheels = Var(
            initialize=4,
            bounds=(1, 10),
            domain=PositiveIntegers,
            units=pyunits.dimensionless,
            doc="Number of rotating paddle wheels in the flocculation basin",
        )

        self.num_paddles_per_wheel = Var(
            initialize=4,
            bounds=(1, 10),
            domain=PositiveIntegers,
            units=pyunits.dimensionless,
            doc="Number of paddles attached to each rotating wheel in the flocculation basin",
        )

        # Build control volume for feed side
        self.control_volume = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type, has_mass_transfer=True
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_enthalpy_transfer=False,
        )

        if self.config.is_isothermal:
            self.control_volume.add_isothermal_assumption()

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=False
        )

        # Add ports
        self.add_inlet_port(name="inlet", block=self.control_volume)
        self.add_outlet_port(name="outlet", block=self.control_volume)

        # Check _phase_component_set for required items
        if ("Liq", "TDS") not in self.config.property_package._phase_component_set:
            raise ConfigurationError(
                "Coagulation-Flocculation model MUST contain ('Liq','TDS') as a component, but "
                "the property package has only specified the following components {}".format(
                    [p for p in self.config.property_package._phase_component_set]
                )
            )
        if ("Liq", "Sludge") not in self.config.property_package._phase_component_set:
            raise ConfigurationError(
                "Coagulation-Flocculation model MUST contain ('Liq','Sludge') as a component, but "
                "the property package has only specified the following components {}".format(
                    [p for p in self.config.property_package._phase_component_set]
                )
            )
        if ("Liq", "TSS") not in self.config.property_package._phase_component_set:
            raise ConfigurationError(
                "Coagulation-Flocculation model MUST contain ('Liq','TSS') as a component, but "
                "the property package has only specified the following components {}".format(
                    [p for p in self.config.property_package._phase_component_set]
                )
            )

        # Constraint for tss loss rate based on measured final turbidity
        self.tss_loss_rate = Var(
            self.flowsheet().config.time,
            initialize=1,
            bounds=(0, 100),
            domain=NonNegativeReals,
            units=units_meta("mass") * units_meta("time") ** -1,
            doc="Mass per time loss rate of TSS based on the measured final turbidity",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Constraint for the loss rate of TSS to be used in mass_transfer_term",
        )
        def eq_tss_loss_rate(self, t):
            tss_out = pyunits.convert(
                self.slope[t] * self.final_turbidity_ntu[t] + self.intercept[t],
                to_units=units_meta("mass") * units_meta("length") ** -3,
            )
            input_rate = self.control_volume.properties_in[t].flow_mass_phase_comp[
                "Liq", "TSS"
            ]
            exit_rate = (
                self.control_volume.properties_out[t].flow_vol_phase["Liq"] * tss_out
            )

            return self.tss_loss_rate[t] == input_rate - exit_rate

        # Constraint for tds gain rate based on 'chemical_doses' and 'chemical_additives'
        if self.config.chemical_additives:
            self.tds_gain_rate = Var(
                self.flowsheet().config.time,
                initialize=0,
                bounds=(0, 100),
                domain=NonNegativeReals,
                units=units_meta("mass") * units_meta("time") ** -1,
                doc="Mass per time gain rate of TDS based on the chemicals added for coagulation",
            )

            @self.Constraint(
                self.flowsheet().config.time,
                doc="Constraint for the loss rate of TSS to be used in mass_transfer_term",
            )
            def eq_tds_gain_rate(self, t):
                sum = 0
                for j in self.config.chemical_additives.keys():
                    chem_dose = pyunits.convert(
                        self.chemical_doses[t, j],
                        to_units=units_meta("mass") * units_meta("length") ** -3,
                    )
                    chem_dose = (
                        chem_dose
                        / self.chemical_mw[j]
                        * self.salt_from_additive_mole_ratio[j]
                        * self.salt_mw[j]
                        * self.control_volume.properties_out[t].flow_vol_phase["Liq"]
                    )
                    sum = sum + chem_dose

                return self.tds_gain_rate[t] == sum

        # Add constraints for mass transfer terms
        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass transfer term",
        )
        def eq_mass_transfer_term(self, t, p, j):
            if (p, j) == ("Liq", "TSS"):
                return (
                    self.control_volume.mass_transfer_term[t, p, j]
                    == -self.tss_loss_rate[t]
                )
            elif (p, j) == ("Liq", "Sludge"):
                return (
                    self.control_volume.mass_transfer_term[t, p, j]
                    == self.tss_loss_rate[t]
                )
            elif (p, j) == ("Liq", "TDS"):
                if self.config.chemical_additives:
                    return (
                        self.control_volume.mass_transfer_term[t, p, j]
                        == self.tds_gain_rate[t]
                    )
                else:
                    return self.control_volume.mass_transfer_term[t, p, j] == 0.0
            else:
                return self.control_volume.mass_transfer_term[t, p, j] == 0.0

        # Constraint for the volume of each rapid mixing basin in series
        #   Mines, R.O., Environmental engineering: Principles
        #       and Practice, 1st Ed, John Wiley & Sons, 2014.
        #       Ch. 6.
        self.rapid_mixing_basin_vol = Var(
            self.flowsheet().config.time,
            initialize=1,
            bounds=(0, 1000),
            domain=NonNegativeReals,
            units=units_meta("length") ** 3,
            doc="Volume of each rapid mixing basin in the series",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Constraint for the volume of each rapid mixing basin",
        )
        def eq_rapid_mixing_basin_vol(self, t):
            flow_rate = pyunits.convert(
                self.control_volume.properties_in[t].flow_vol_phase["Liq"],
                to_units=units_meta("length") ** 3 / pyunits.s,
            )

            return (
                self.rapid_mixing_basin_vol[t]
                == flow_rate * self.rapid_mixing_retention_time[t]
            )

        # Constraint for the power usage of the rapid mixers
        #   Mines, R.O., Environmental engineering: Principles
        #       and Practice, 1st Ed, John Wiley & Sons, 2014.
        #       Ch. 6.
        self.rapid_mixing_power = Var(
            self.flowsheet().config.time,
            initialize=0.01,
            bounds=(0, 100),
            domain=NonNegativeReals,
            units=pyunits.kW,
            doc="Power usage of the rapid mixing basins in kW",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Constraint for the power usage of the rapid mixing basins",
        )
        def eq_rapid_mixing_power(self, t):
            vel_grad = pyunits.convert(
                self.rapid_mixing_vel_grad[t], to_units=units_meta("time") ** -1
            )
            power_usage = pyunits.convert(
                vel_grad**2
                * self.control_volume.properties_out[t].visc_d_phase["Liq"]
                * self.rapid_mixing_basin_vol[t]
                * self.num_rapid_mixing_basins,
                to_units=pyunits.kW,
            )

            return self.rapid_mixing_power[t] == power_usage

        # Constraint for the volume of the flocculation basin
        #   Mines, R.O., Environmental engineering: Principles
        #       and Practice, 1st Ed, John Wiley & Sons, 2014.
        #       Ch. 6.
        self.floc_basin_vol = Var(
            self.flowsheet().config.time,
            initialize=10,
            bounds=(0, 10000),
            domain=NonNegativeReals,
            units=units_meta("length") ** 3,
            doc="Volume of the flocculation basin",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Constraint for the volume of the flocculation basin",
        )
        def eq_floc_basin_vol(self, t):
            flow_rate = pyunits.convert(
                self.control_volume.properties_in[t].flow_vol_phase["Liq"],
                to_units=units_meta("length") ** 3 / pyunits.s,
            )

            return self.floc_basin_vol[t] == flow_rate * self.floc_retention_time[t]

        # Constraint for the velocity of the paddle wheels
        #   Mines, R.O., Environmental engineering: Principles
        #       and Practice, 1st Ed, John Wiley & Sons, 2014.
        #       Ch. 6.
        self.floc_wheel_speed = Var(
            self.flowsheet().config.time,
            initialize=1,
            bounds=(0, 100),
            domain=NonNegativeReals,
            units=units_meta("length") * units_meta("time") ** -1,
            doc="Velocity of the wheels in the flocculation basin",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Constraint for the velocity of the wheels in the flocculation basin",
        )
        def eq_floc_wheel_speed(self, t):
            wheel_rate = pyunits.convert(
                Constants.pi
                * self.single_paddle_length
                * self.paddle_rotational_speed[t],
                to_units=units_meta("length") * units_meta("time") ** -1,
            )

            return self.floc_wheel_speed[t] == wheel_rate

        # Constraint for the power usage of the flocculation mixer
        #   Mines, R.O., Environmental engineering: Principles
        #       and Practice, 1st Ed, John Wiley & Sons, 2014.
        #       Ch. 6.
        self.flocculation_power = Var(
            self.flowsheet().config.time,
            initialize=0.5,
            bounds=(0, 100),
            domain=NonNegativeReals,
            units=pyunits.kW,
            doc="Power usage of the flocculation basin in kW",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Constraint for the power usage of the flocculation basin",
        )
        def eq_flocculation_power(self, t):
            total_area = pyunits.convert(
                self.single_paddle_width
                * self.single_paddle_length
                * self.num_paddle_wheels
                * self.num_paddles_per_wheel,
                to_units=units_meta("length") ** 2,
            )
            power_usage = pyunits.convert(
                0.5
                * self.paddle_drag_coef[t]
                * total_area
                * self.control_volume.properties_out[t].dens_mass_phase["Liq"]
                * self.vel_fraction**3
                * self.floc_wheel_speed[t] ** 3,
                to_units=pyunits.kW,
            )

            return self.flocculation_power[t] == power_usage

        self.total_power = Var(
            self.flowsheet().config.time,
            initialize=0.5,
            bounds=(0, 100),
            domain=NonNegativeReals,
            units=pyunits.kW,
            doc="Power usage of the full unit model in kW",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Constraint for the power usage of the full unit model",
        )
        def eq_total_power(self, t):
            return (
                self.total_power[t]
                == self.flocculation_power[t] + self.rapid_mixing_power[t]
            )

    # Return a scalar expression for the inlet concentration of TSS
    def compute_inlet_tss_mass_concentration(self, t):
        """
        Function to generate an expression that would represent the mass
        concentration of TSS at the inlet port of the unit. Inlet ports
        are generally established upstream, but this will be useful for
        establishing the inlet TSS when an upstream TSS is unknown. This
        level of inlet TSS is based off of measurements made of Turbidity
        during the Jar Test.

        Keyword Arguments:
            self : this unit model object
            t : time index on the flowsheet

        Returns: Expression

        Recover the numeric value by using 'value(Expression)'
        """
        units_meta = self.config.property_package.get_metadata().get_derived_units
        return pyunits.convert(
            self.slope[t] * self.initial_turbidity_ntu[t] + self.intercept[t],
            to_units=units_meta("mass") * units_meta("length") ** -3,
        )

    # Return a scale expression for the inlet mass flow rate of TSS
    def compute_inlet_tss_mass_flow(self, t):
        """
        Function to generate an expression that would represent the mass
        flow rate of TSS at the inlet port of the unit. Inlet ports
        are generally established upstream, but this will be useful for
        establishing the inlet TSS when an upstream TSS is unknown. This
        level of inlet TSS is based off of measurements made of Turbidity
        during the Jar Test.

        Keyword Arguments:
            self : this unit model object
            t : time index on the flowsheet

        Returns: Expression

        Recover the numeric value by using 'value(Expression)'
        """
        return self.control_volume.properties_in[t].flow_vol_phase[
            "Liq"
        ] * self.compute_inlet_tss_mass_concentration(t)

    # Function to automate fixing of the Turbidity v TSS relation params to defaults
    def fix_tss_turbidity_relation_defaults(self):
        self.slope.fix()
        self.intercept.fix()

    # initialize method
    def initialize_build(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize holdup block
        flags = self.control_volume.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )
        init_log.info_high("Initialization Step 1 Complete.")
        # ---------------------------------------------------------------------

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        self.control_volume.release_state(flags, outlvl + 1)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        units_meta = self.config.property_package.get_metadata().get_derived_units

        # scaling factors for turbidity relationship
        #       Supressing warning (these factors are not very important)
        if iscale.get_scaling_factor(self.slope) is None:
            sf = iscale.get_scaling_factor(self.slope, default=1, warning=False)
            iscale.set_scaling_factor(self.slope, sf)
        if iscale.get_scaling_factor(self.intercept) is None:
            sf = iscale.get_scaling_factor(self.intercept, default=1, warning=False)
            iscale.set_scaling_factor(self.intercept, sf)

        # scaling factors for turbidity measurements and chemical doses
        #       Supressing warning
        if iscale.get_scaling_factor(self.initial_turbidity_ntu) is None:
            sf = iscale.get_scaling_factor(
                self.initial_turbidity_ntu, default=1, warning=False
            )
            iscale.set_scaling_factor(self.initial_turbidity_ntu, sf)
        if iscale.get_scaling_factor(self.final_turbidity_ntu) is None:
            sf = iscale.get_scaling_factor(
                self.final_turbidity_ntu, default=1, warning=False
            )
            iscale.set_scaling_factor(self.final_turbidity_ntu, sf)
        if iscale.get_scaling_factor(self.chemical_doses) is None:
            sf = iscale.get_scaling_factor(
                self.chemical_doses, default=1, warning=False
            )
            iscale.set_scaling_factor(self.chemical_doses, sf)

        # scaling factors for power usage in rapid mixing
        #       Display warning
        if iscale.get_scaling_factor(self.rapid_mixing_retention_time) is None:
            sf = iscale.get_scaling_factor(
                self.rapid_mixing_retention_time, default=1e-1, warning=True
            )
            iscale.set_scaling_factor(self.rapid_mixing_retention_time, sf)
        if iscale.get_scaling_factor(self.num_rapid_mixing_basins) is None:
            sf = iscale.get_scaling_factor(
                self.num_rapid_mixing_basins, default=1, warning=True
            )
            iscale.set_scaling_factor(self.num_rapid_mixing_basins, sf)
        if iscale.get_scaling_factor(self.rapid_mixing_vel_grad) is None:
            sf = iscale.get_scaling_factor(
                self.rapid_mixing_vel_grad, default=1e-2, warning=True
            )
            iscale.set_scaling_factor(self.rapid_mixing_vel_grad, sf)
        if iscale.get_scaling_factor(self.floc_retention_time) is None:
            sf = iscale.get_scaling_factor(
                self.floc_retention_time, default=1e-3, warning=True
            )
            iscale.set_scaling_factor(self.floc_retention_time, sf)
        if iscale.get_scaling_factor(self.single_paddle_length) is None:
            sf = iscale.get_scaling_factor(
                self.single_paddle_length, default=1, warning=True
            )
            iscale.set_scaling_factor(self.single_paddle_length, sf)
        if iscale.get_scaling_factor(self.single_paddle_width) is None:
            sf = iscale.get_scaling_factor(
                self.single_paddle_width, default=1, warning=True
            )
            iscale.set_scaling_factor(self.single_paddle_width, sf)
        if iscale.get_scaling_factor(self.paddle_rotational_speed) is None:
            sf = iscale.get_scaling_factor(
                self.paddle_rotational_speed, default=10, warning=True
            )
            iscale.set_scaling_factor(self.paddle_rotational_speed, sf)
        if iscale.get_scaling_factor(self.paddle_drag_coef) is None:
            sf = iscale.get_scaling_factor(
                self.paddle_drag_coef, default=1, warning=True
            )
            iscale.set_scaling_factor(self.paddle_drag_coef, sf)
        if iscale.get_scaling_factor(self.vel_fraction) is None:
            sf = iscale.get_scaling_factor(self.vel_fraction, default=1, warning=True)
            iscale.set_scaling_factor(self.vel_fraction, sf)
        if iscale.get_scaling_factor(self.num_paddle_wheels) is None:
            sf = iscale.get_scaling_factor(
                self.num_paddle_wheels, default=1, warning=True
            )
            iscale.set_scaling_factor(self.num_paddle_wheels, sf)
        if iscale.get_scaling_factor(self.num_paddles_per_wheel) is None:
            sf = iscale.get_scaling_factor(
                self.num_paddles_per_wheel, default=1, warning=True
            )
            iscale.set_scaling_factor(self.num_paddles_per_wheel, sf)

        # set scaling for tss_loss_rate
        if iscale.get_scaling_factor(self.tss_loss_rate) is None:
            sf = 0
            for t in self.control_volume.properties_in:
                sf += value(
                    self.control_volume.properties_in[t].flow_mass_phase_comp[
                        "Liq", "TSS"
                    ]
                )
            sf = sf / len(self.control_volume.properties_in)
            if sf < 0.01:
                sf = 0.01
            iscale.set_scaling_factor(self.tss_loss_rate, 1 / sf)

            for ind, c in self.eq_tss_loss_rate.items():
                iscale.constraint_scaling_transform(c, 1 / sf)

        # set scaling for tds_gain_rate
        if self.config.chemical_additives:
            if iscale.get_scaling_factor(self.tds_gain_rate) is None:
                sf = 0
                for t in self.control_volume.properties_in:
                    sum = 0
                    for j in self.config.chemical_additives.keys():
                        chem_dose = pyunits.convert(
                            self.chemical_doses[t, j],
                            to_units=units_meta("mass") * units_meta("length") ** -3,
                        )
                        chem_dose = (
                            chem_dose
                            / self.chemical_mw[j]
                            * self.salt_from_additive_mole_ratio[j]
                            * self.salt_mw[j]
                            * self.control_volume.properties_in[t].flow_vol_phase["Liq"]
                        )
                        sum = sum + chem_dose
                    sf += value(sum)
                sf = sf / len(self.control_volume.properties_in)
                if sf < 0.001:
                    sf = 0.001
                iscale.set_scaling_factor(self.tds_gain_rate, 1 / sf)

                for ind, c in self.eq_tds_gain_rate.items():
                    iscale.constraint_scaling_transform(c, 1 / sf)

        # set scaling for mass transfer terms
        for ind, c in self.eq_mass_transfer_term.items():
            if ind[2] == "TDS":
                if self.config.chemical_additives:
                    sf = iscale.get_scaling_factor(self.tds_gain_rate)
                else:
                    sf = 1
            elif ind[2] == "TSS":
                sf = iscale.get_scaling_factor(self.tss_loss_rate)
            elif ind[2] == "Sludge":
                sf = iscale.get_scaling_factor(self.tss_loss_rate)
            else:
                sf = 1
            iscale.constraint_scaling_transform(c, sf)
            iscale.set_scaling_factor(self.control_volume.mass_transfer_term[ind], sf)

        # set scaling for rapid_mixing_basin_vol
        if iscale.get_scaling_factor(self.rapid_mixing_basin_vol[t]) is None:
            sf1 = 0
            sf2 = 0
            for t in self.control_volume.properties_out:
                sf1 += iscale.get_scaling_factor(
                    self.control_volume.properties_out[t].flow_vol_phase["Liq"]
                )
            sf2 = iscale.get_scaling_factor(self.rapid_mixing_retention_time)
            sf1 = sf1 / len(self.control_volume.properties_in)
            sf = sf1 * sf2
            iscale.set_scaling_factor(self.rapid_mixing_basin_vol, sf)

            for ind, c in self.eq_rapid_mixing_basin_vol.items():
                iscale.constraint_scaling_transform(c, sf)

        # set scaling for rapid_mixing_power
        if iscale.get_scaling_factor(self.rapid_mixing_power[t]) is None:
            sf1 = 0
            sf2 = 0
            for t in self.control_volume.properties_out:
                sf1 += iscale.get_scaling_factor(
                    self.control_volume.properties_out[t].visc_d_phase["Liq"]
                )
            sf2 = iscale.get_scaling_factor(self.rapid_mixing_vel_grad)
            sf3 = iscale.get_scaling_factor(self.rapid_mixing_basin_vol)
            sf4 = iscale.get_scaling_factor(self.num_rapid_mixing_basins)
            sf1 = sf1 / len(self.control_volume.properties_in)
            sf = sf1 * sf2**2 * sf3 * sf4
            iscale.set_scaling_factor(self.rapid_mixing_power, sf)

            for ind, c in self.eq_rapid_mixing_power.items():
                iscale.constraint_scaling_transform(c, sf)

        # set scaling for floc_basin_vol
        if iscale.get_scaling_factor(self.floc_basin_vol[t]) is None:
            sf1 = 0
            sf2 = 0
            for t in self.control_volume.properties_out:
                sf1 += iscale.get_scaling_factor(
                    self.control_volume.properties_out[t].flow_vol_phase["Liq"]
                )
            sf2 = iscale.get_scaling_factor(self.floc_retention_time)
            sf1 = sf1 / len(self.control_volume.properties_in)
            sf = sf1 * sf2
            iscale.set_scaling_factor(self.floc_basin_vol, sf)

            for ind, c in self.eq_floc_basin_vol.items():
                iscale.constraint_scaling_transform(c, sf)

        # set scaling for floc_wheel_speed
        if iscale.get_scaling_factor(self.floc_wheel_speed[t]) is None:
            sf1 = iscale.get_scaling_factor(self.paddle_rotational_speed)
            sf2 = iscale.get_scaling_factor(self.single_paddle_length)
            sf = sf1 * sf2 * Constants.pi / 10
            iscale.set_scaling_factor(self.floc_wheel_speed, sf)

            for ind, c in self.eq_floc_wheel_speed.items():
                iscale.constraint_scaling_transform(c, sf)

        # set scaling for flocculation_power
        if iscale.get_scaling_factor(self.flocculation_power[t]) is None:
            sf1 = iscale.get_scaling_factor(self.floc_wheel_speed)
            sf2 = iscale.get_scaling_factor(self.vel_fraction)
            sf3 = 0
            for t in self.control_volume.properties_out:
                sf3 += iscale.get_scaling_factor(
                    self.control_volume.properties_out[t].dens_mass_phase["Liq"]
                )
            sf3 = sf3 / len(self.control_volume.properties_in)
            sf4 = iscale.get_scaling_factor(self.single_paddle_length)
            sf5 = iscale.get_scaling_factor(self.single_paddle_width)
            sf6 = iscale.get_scaling_factor(self.num_paddle_wheels)
            sf7 = iscale.get_scaling_factor(self.num_paddles_per_wheel)
            sf8 = iscale.get_scaling_factor(self.paddle_drag_coef)
            sf = 0.5 * sf8 * (sf4 * sf5 * sf6 * sf7) * sf3 * sf2**3 * sf1**3 * 500
            iscale.set_scaling_factor(self.flocculation_power, sf)

            for ind, c in self.eq_flocculation_power.items():
                iscale.constraint_scaling_transform(c, sf)

        # set scaling for total_power
        if iscale.get_scaling_factor(self.total_power[t]) is None:
            sf1 = iscale.get_scaling_factor(self.flocculation_power)
            sf2 = iscale.get_scaling_factor(self.rapid_mixing_power)
            sf = (sf1 + sf2) / 2
            iscale.set_scaling_factor(self.total_power, sf)

            for ind, c in self.eq_total_power.items():
                iscale.constraint_scaling_transform(c, sf)

        # set scaling factors for control_volume.properties_in based on control_volume.properties_out
        for t in self.control_volume.properties_in:
            if (
                iscale.get_scaling_factor(
                    self.control_volume.properties_in[t].dens_mass_phase
                )
                is None
            ):
                sf = iscale.get_scaling_factor(
                    self.control_volume.properties_out[t].dens_mass_phase
                )
                iscale.set_scaling_factor(
                    self.control_volume.properties_in[t].dens_mass_phase, sf
                )

            if (
                iscale.get_scaling_factor(
                    self.control_volume.properties_in[t].flow_mass_phase_comp
                )
                is None
            ):
                for ind in self.control_volume.properties_in[t].flow_mass_phase_comp:
                    sf = iscale.get_scaling_factor(
                        self.control_volume.properties_out[t].flow_mass_phase_comp[ind]
                    )
                    iscale.set_scaling_factor(
                        self.control_volume.properties_in[t].flow_mass_phase_comp[ind],
                        sf,
                    )

            if (
                iscale.get_scaling_factor(
                    self.control_volume.properties_in[t].mass_frac_phase_comp
                )
                is None
            ):
                for ind in self.control_volume.properties_in[t].mass_frac_phase_comp:
                    sf = iscale.get_scaling_factor(
                        self.control_volume.properties_out[t].mass_frac_phase_comp[ind]
                    )
                    iscale.set_scaling_factor(
                        self.control_volume.properties_in[t].mass_frac_phase_comp[ind],
                        sf,
                    )

            if (
                iscale.get_scaling_factor(
                    self.control_volume.properties_in[t].flow_vol_phase
                )
                is None
            ):
                for ind in self.control_volume.properties_in[t].flow_vol_phase:
                    sf = iscale.get_scaling_factor(
                        self.control_volume.properties_out[t].flow_vol_phase[ind]
                    )
                    iscale.set_scaling_factor(
                        self.control_volume.properties_in[t].flow_vol_phase[ind], sf
                    )

        # update scaling for control_volume.properties_out
        for t in self.control_volume.properties_out:
            if (
                iscale.get_scaling_factor(
                    self.control_volume.properties_out[t].dens_mass_phase
                )
                is None
            ):
                iscale.set_scaling_factor(
                    self.control_volume.properties_out[t].dens_mass_phase, 1e-3
                )
            if (
                iscale.get_scaling_factor(
                    self.control_volume.properties_out[t].visc_d_phase
                )
                is None
            ):
                iscale.set_scaling_factor(
                    self.control_volume.properties_out[t].visc_d_phase, 1e3
                )

            # need to update scaling factors for TSS, Sludge, and TDS to account for the
            #   expected change in their respective values from the loss/gain rates
            for ind in self.control_volume.properties_out[t].flow_mass_phase_comp:
                if ind[1] == "TSS":
                    sf_og = iscale.get_scaling_factor(
                        self.control_volume.properties_out[t].flow_mass_phase_comp[ind]
                    )
                    sf_new = iscale.get_scaling_factor(self.tss_loss_rate)
                    iscale.set_scaling_factor(
                        self.control_volume.properties_out[t].flow_mass_phase_comp[ind],
                        100 * sf_new * (sf_new / sf_og),
                    )
                if ind[1] == "Sludge":
                    sf_og = iscale.get_scaling_factor(
                        self.control_volume.properties_out[t].flow_mass_phase_comp[ind]
                    )
                    sf_new = iscale.get_scaling_factor(self.tss_loss_rate)
                    iscale.set_scaling_factor(
                        self.control_volume.properties_out[t].flow_mass_phase_comp[ind],
                        100 * sf_new * (sf_new / sf_og),
                    )

            for ind in self.control_volume.properties_out[t].mass_frac_phase_comp:
                if ind[1] == "TSS":
                    sf_og = iscale.get_scaling_factor(
                        self.control_volume.properties_out[t].mass_frac_phase_comp[ind]
                    )
                    sf_new = iscale.get_scaling_factor(self.tss_loss_rate)
                    iscale.set_scaling_factor(
                        self.control_volume.properties_out[t].mass_frac_phase_comp[ind],
                        100 * sf_new * (sf_new / sf_og),
                    )
                if ind[1] == "Sludge":
                    sf_og = iscale.get_scaling_factor(
                        self.control_volume.properties_out[t].mass_frac_phase_comp[ind]
                    )
                    sf_new = iscale.get_scaling_factor(self.tss_loss_rate)
                    iscale.set_scaling_factor(
                        self.control_volume.properties_out[t].mass_frac_phase_comp[ind],
                        100 * sf_new * (sf_new / sf_og),
                    )

    def _get_performance_contents(self, time_point=0):
        t = time_point
        return {
            "vars": {
                "Total Power Usage  (kW)": self.total_power[t],
                "Rapid Mixing Power (kW)": self.rapid_mixing_power[t],
                "Flocc Mixing Power (kW)": self.flocculation_power[t],
            },
            "exprs": {},
            "params": {},
        }



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Based on flowsheet from:

Flores-Alsina X., Gernaey K.V. and Jeppsson, U. "Benchmarking biological
nutrient removal in wastewater treatment plants: influence of mathematical model
assumptions", 2012, Wat. Sci. Tech., Vol. 65 No. 8, pp. 1496-1505
"""

# Some more information about this module
__author__ = "Chenyu Wang"

import pyomo.environ as pyo
from pyomo.network import Arc, SequentialDecomposition

from idaes.core import FlowsheetBlock
from idaes.models.unit_models import (
    CSTR,
    Feed,
    Mixer,
    Separator,
    PressureChanger,
    Product,
)
from idaes.models.unit_models.separator import SplittingType
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from idaes.core.util.tables import (
    create_stream_table_dataframe,
    stream_table_dataframe_to_string,
)

from watertap.unit_models.cstr_injection import CSTR_Injection
from watertap.property_models.activated_sludge.modified_asm2d_properties import (
    ModifiedASM2dParameterBlock,
)
from watertap.property_models.activated_sludge.modified_asm2d_reactions import (
    ModifiedASM2dReactionParameterBlock,
)
from idaes.models.unit_models.mixer import MomentumMixingType
from watertap.core.util.initialization import check_solve

# Set up logger
_log = idaeslog.getLogger(__name__)


def build_flowsheet():
    m = pyo.ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ModifiedASM2dParameterBlock()
    m.fs.rxn_props = ModifiedASM2dReactionParameterBlock(property_package=m.fs.props)

    # Feed water stream
    m.fs.FeedWater = Feed(property_package=m.fs.props)
    # Mixer for feed water and recycled sludge
    m.fs.MX1 = Mixer(
        property_package=m.fs.props,
        inlet_list=["feed_water", "recycle"],
        momentum_mixing_type=MomentumMixingType.equality,
    )
    # First reactor (anaerobic) - standard CSTR
    m.fs.R1 = CSTR(property_package=m.fs.props, reaction_package=m.fs.rxn_props)
    # First reactor (anaerobic) - standard CSTR
    m.fs.R2 = CSTR(property_package=m.fs.props, reaction_package=m.fs.rxn_props)
    # Third reactor (anoxic) - standard CSTR
    m.fs.R3 = CSTR(property_package=m.fs.props, reaction_package=m.fs.rxn_props)
    # Forth reactor (anoxic) - standard CSTR
    m.fs.R4 = CSTR(property_package=m.fs.props, reaction_package=m.fs.rxn_props)
    # Fifth reactor (aerobic) - CSTR with injection
    m.fs.R5 = CSTR_Injection(
        property_package=m.fs.props, reaction_package=m.fs.rxn_props
    )
    # Sixth reactor (aerobic) - CSTR with injection
    m.fs.R6 = CSTR_Injection(
        property_package=m.fs.props, reaction_package=m.fs.rxn_props
    )
    # Seventh reactor (aerobic) - CSTR with injection
    m.fs.R7 = CSTR_Injection(
        property_package=m.fs.props, reaction_package=m.fs.rxn_props
    )
    m.fs.SP1 = Separator(
        property_package=m.fs.props, outlet_list=["underflow", "overflow"]
    )
    # Clarifier
    # TODO: Replace with more detailed model when available
    m.fs.CL1 = Separator(
        property_package=m.fs.props,
        outlet_list=["underflow", "effluent"],
        split_basis=SplittingType.componentFlow,
    )
    # Mixing sludge recycle and R5 underflow
    m.fs.MX2 = Mixer(
        property_package=m.fs.props,
        inlet_list=["reactor", "clarifier"],
        momentum_mixing_type=MomentumMixingType.equality,
    )
    # Sludge separator
    m.fs.SP2 = Separator(property_package=m.fs.props, outlet_list=["waste", "recycle"])
    # Product Blocks
    m.fs.Treated = Product(property_package=m.fs.props)
    m.fs.Sludge = Product(property_package=m.fs.props)
    # Recycle pressure changer - use a simple isothermal unit for now
    m.fs.P1 = PressureChanger(property_package=m.fs.props)

    m.fs.mixers = (m.fs.MX1, m.fs.MX2)

    # Link units
    m.fs.stream1 = Arc(source=m.fs.FeedWater.outlet, destination=m.fs.MX1.feed_water)
    m.fs.stream2 = Arc(source=m.fs.MX1.outlet, destination=m.fs.R1.inlet)
    m.fs.stream3 = Arc(source=m.fs.R1.outlet, destination=m.fs.R2.inlet)
    m.fs.stream4 = Arc(source=m.fs.R2.outlet, destination=m.fs.MX2.reactor)
    m.fs.stream5 = Arc(source=m.fs.MX2.outlet, destination=m.fs.R3.inlet)
    m.fs.stream6 = Arc(source=m.fs.R3.outlet, destination=m.fs.R4.inlet)
    m.fs.stream7 = Arc(source=m.fs.R4.outlet, destination=m.fs.R5.inlet)
    m.fs.stream8 = Arc(source=m.fs.R5.outlet, destination=m.fs.R6.inlet)
    m.fs.stream9 = Arc(source=m.fs.R6.outlet, destination=m.fs.R7.inlet)
    m.fs.stream10 = Arc(source=m.fs.R7.outlet, destination=m.fs.SP1.inlet)
    m.fs.stream11 = Arc(source=m.fs.SP1.overflow, destination=m.fs.CL1.inlet)
    m.fs.stream12 = Arc(source=m.fs.SP1.underflow, destination=m.fs.MX2.clarifier)
    m.fs.stream13 = Arc(source=m.fs.CL1.effluent, destination=m.fs.Treated.inlet)
    m.fs.stream14 = Arc(source=m.fs.CL1.underflow, destination=m.fs.SP2.inlet)

    m.fs.stream15 = Arc(source=m.fs.SP2.waste, destination=m.fs.Sludge.inlet)
    m.fs.stream16 = Arc(source=m.fs.SP2.recycle, destination=m.fs.P1.inlet)

    m.fs.stream17 = Arc(source=m.fs.P1.outlet, destination=m.fs.MX1.recycle)
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Oxygen concentration in reactors 3 and 4 is governed by mass transfer
    # Add additional parameter and constraints
    m.fs.R5.KLa = pyo.Var(
        initialize=240,
        units=pyo.units.hour**-1,
        doc="Lumped mass transfer coefficient for oxygen",
    )
    m.fs.R6.KLa = pyo.Var(
        initialize=240,
        units=pyo.units.hour**-1,
        doc="Lumped mass transfer coefficient for oxygen",
    )
    m.fs.R7.KLa = pyo.Var(
        initialize=84,
        units=pyo.units.hour**-1,
        doc="Lumped mass transfer coefficient for oxygen",
    )
    m.fs.S_O_eq = pyo.Param(
        default=8e-3,
        units=pyo.units.kg / pyo.units.m**3,
        mutable=True,
        doc="Dissolved oxygen concentration at equilibrium",
    )

    @m.fs.R5.Constraint(m.fs.time, doc="Mass transfer constraint for R3")
    def mass_transfer_R5(self, t):
        return pyo.units.convert(
            m.fs.R5.injection[t, "Liq", "S_O2"], to_units=pyo.units.kg / pyo.units.hour
        ) == (
            m.fs.R5.KLa
            * m.fs.R5.volume[t]
            * (m.fs.S_O_eq - m.fs.R5.outlet.conc_mass_comp[t, "S_O2"])
        )

    @m.fs.R6.Constraint(m.fs.time, doc="Mass transfer constraint for R4")
    def mass_transfer_R6(self, t):
        return pyo.units.convert(
            m.fs.R6.injection[t, "Liq", "S_O2"], to_units=pyo.units.kg / pyo.units.hour
        ) == (
            m.fs.R6.KLa
            * m.fs.R6.volume[t]
            * (m.fs.S_O_eq - m.fs.R6.outlet.conc_mass_comp[t, "S_O2"])
        )

    @m.fs.R7.Constraint(m.fs.time, doc="Mass transfer constraint for R4")
    def mass_transfer_R7(self, t):
        return pyo.units.convert(
            m.fs.R7.injection[t, "Liq", "S_O2"], to_units=pyo.units.kg / pyo.units.hour
        ) == (
            m.fs.R7.KLa
            * m.fs.R7.volume[t]
            * (m.fs.S_O_eq - m.fs.R7.outlet.conc_mass_comp[t, "S_O2"])
        )

    # Feed Water Conditions
    print(f"DOF before feed: {degrees_of_freedom(m)}")
    m.fs.FeedWater.flow_vol.fix(18446 * pyo.units.m**3 / pyo.units.day)
    m.fs.FeedWater.temperature.fix(298.15 * pyo.units.K)
    m.fs.FeedWater.pressure.fix(1 * pyo.units.atm)
    m.fs.FeedWater.conc_mass_comp[0, "S_O2"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_F"].fix(30 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_A"].fix(20 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NH4"].fix(16 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NO3"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_PO4"].fix(3.6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_I"].fix(30 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_N2"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_I"].fix(25 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_S"].fix(125 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_H"].fix(30 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_PAO"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_PP"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_PHA"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_AUT"].fix(1e-6 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.properties[0].conc_mass_comp["S_IC"].fix(
        0.07899 * pyo.units.kg / pyo.units.m**3
    )
    m.fs.FeedWater.properties[0].conc_mass_comp["S_K"].fix(
        1e-6 * pyo.units.g / pyo.units.m**3
    )
    m.fs.FeedWater.properties[0].conc_mass_comp["S_Mg"].fix(
        1e-6 * pyo.units.g / pyo.units.m**3
    )

    # Reactor sizing
    m.fs.R1.volume.fix(1000 * pyo.units.m**3)
    m.fs.R2.volume.fix(1000 * pyo.units.m**3)
    m.fs.R3.volume.fix(1000 * pyo.units.m**3)
    m.fs.R4.volume.fix(1000 * pyo.units.m**3)
    m.fs.R5.volume.fix(1333 * pyo.units.m**3)
    m.fs.R6.volume.fix(1333 * pyo.units.m**3)
    m.fs.R7.volume.fix(1333 * pyo.units.m**3)

    # Injection rates to Reactions 3, 4 and 5
    for j in m.fs.props.component_list:
        if j != "S_O2":
            # All components except S_O have no injection
            m.fs.R5.injection[:, :, j].fix(0)
            m.fs.R6.injection[:, :, j].fix(0)
            m.fs.R7.injection[:, :, j].fix(0)
    # Then set injections rates for O2
    m.fs.R5.outlet.conc_mass_comp[:, "S_O2"].fix(1.91e-3)
    m.fs.R6.outlet.conc_mass_comp[:, "S_O2"].fix(2.60e-3)
    m.fs.R7.outlet.conc_mass_comp[:, "S_O2"].fix(3.20e-3)

    # Set fraction of outflow from reactor 5 that goes to recycle
    m.fs.SP1.split_fraction[:, "underflow"].fix(0.60)

    # Clarifier
    # TODO: Update once more detailed model available
    m.fs.CL1.split_fraction[0, "effluent", "H2O"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_A"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_F"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_I"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_N2"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_NH4"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_NO3"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_O2"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_PO4"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_IC"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_K"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "S_Mg"].fix(0.49986)
    m.fs.CL1.split_fraction[0, "effluent", "X_AUT"].fix(0.022117)
    m.fs.CL1.split_fraction[0, "effluent", "X_H"].fix(0.021922)
    m.fs.CL1.split_fraction[0, "effluent", "X_I"].fix(0.021715)
    m.fs.CL1.split_fraction[0, "effluent", "X_PAO"].fix(0.022)
    m.fs.CL1.split_fraction[0, "effluent", "X_PHA"].fix(0.02147)
    m.fs.CL1.split_fraction[0, "effluent", "X_PP"].fix(0.02144)
    m.fs.CL1.split_fraction[0, "effluent", "X_S"].fix(0.02221)

    # Sludge purge separator
    m.fs.SP2.split_fraction[:, "recycle"].fix(0.97955)

    # Outlet pressure from recycle pump
    m.fs.P1.outlet.pressure.fix(101325)

    # Check degrees of freedom
    for mx in m.fs.mixers:
        mx.pressure_equality_constraints[0.0, 2].deactivate()
    print(f"DOF before feed: {degrees_of_freedom(m)}")
    assert degrees_of_freedom(m) == 0

    def scale_variables(m):
        for var in m.fs.component_data_objects(pyo.Var, descend_into=True):
            if "flow_vol" in var.name:
                iscale.set_scaling_factor(var, 1e1)
            if "temperature" in var.name:
                iscale.set_scaling_factor(var, 1e-1)
            if "pressure" in var.name:
                iscale.set_scaling_factor(var, 1e-4)
            if "conc_mass_comp" in var.name:
                iscale.set_scaling_factor(var, 1e1)

    for unit in ("R1", "R2", "R3", "R4", "R5", "R6", "R7"):
        block = getattr(m.fs, unit)
        iscale.set_scaling_factor(
            block.control_volume.reactions[0.0].rate_expression, 1e3
        )
        iscale.set_scaling_factor(block.cstr_performance_eqn, 1e3)
        iscale.set_scaling_factor(
            block.control_volume.rate_reaction_stoichiometry_constraint, 1e3
        )
        iscale.set_scaling_factor(block.control_volume.material_balances, 1e3)

    # Apply scaling
    scale_variables(m)
    iscale.calculate_scaling_factors(m)

    # Initialize flowsheet
    # Apply sequential decomposition - 1 iteration should suffice
    seq = SequentialDecomposition()
    seq.options.select_tear_method = "heuristic"
    seq.options.tear_method = "Wegstein"
    seq.options.iterLim = 1

    G = seq.create_graph(m)

    tear_guesses = {
        "flow_vol": {0: 1.1},
        "conc_mass_comp": {
            (0, "S_A"): 0.018,
            (0, "S_F"): 0.0006,
            (0, "S_I"): 0.03,
            (0, "S_N2"): 1e-9,
            (0, "S_NH4"): 0.008,
            (0, "S_NO3"): 1e-9,
            (0, "S_O2"): 0.002,
            (0, "S_PO4"): 0.003,
            (0, "S_K"): 1e-9,
            (0, "S_Mg"): 1e-9,
            (0, "S_IC"): 0.1,
            (0, "X_AUT"): 1e-9,
            (0, "X_H"): 1.3,
            (0, "X_I"): 0.38,
            (0, "X_PAO"): 1e-9,
            (0, "X_PHA"): 1e-9,
            (0, "X_PP"): 1e-9,
            (0, "X_S"): 0.04,
        },
        "temperature": {0: 298.15},
        "pressure": {0: 101325},
    }

    # Pass the tear_guess to the SD tool
    seq.set_guesses_for(m.fs.R3.inlet, tear_guesses)

    def function(unit):
        unit.initialize(outlvl=idaeslog.INFO, optarg={"bound_push": 1e-2})

    seq.run(m, function)
    for mx in m.fs.mixers:
        mx.pressure_equality_constraints[0.0, 2].deactivate()
    assert degrees_of_freedom(m) == 0
    # m.fs.MX2.pressure_equality_constraints[0.0, 2].deactivate()

    solver = get_solver()
    results = solver.solve(m, tee=False)
    check_solve(results, checkpoint="closing recycle", logger=_log, fail_flag=True)

    # Switch to fixed KLa in R3 and R4 (S_O concentration is controlled in R5)
    m.fs.R5.KLa.fix(240)
    m.fs.R6.KLa.fix(240)
    m.fs.R7.KLa.fix(84)
    m.fs.R5.outlet.conc_mass_comp[:, "S_O2"].unfix()
    m.fs.R6.outlet.conc_mass_comp[:, "S_O2"].unfix()
    m.fs.R7.outlet.conc_mass_comp[:, "S_O2"].unfix()

    # Resolve with controls in place
    results = solver.solve(m, tee=False)

    pyo.assert_optimal_termination(results)
    check_solve(
        results,
        checkpoint="re-solve with controls in place",
        logger=_log,
        fail_flag=True,
    )

    return m, results


if __name__ == "__main__":
    # This method builds and runs a steady state activated sludge
    # flowsheet.
    m, results = build_flowsheet()

    stream_table = create_stream_table_dataframe(
        {
            "Feed": m.fs.FeedWater.outlet,
            "Mix": m.fs.R1.inlet,
            "R1": m.fs.R1.outlet,
            "R2": m.fs.R2.outlet,
            "R3 inlet": m.fs.R3.inlet,
            "R3": m.fs.R3.outlet,
            "R4": m.fs.R4.outlet,
            "R5": m.fs.R5.outlet,
            "R6": m.fs.R6.outlet,
            "R7": m.fs.R7.outlet,
        },
        time_point=0,
    )
    print(stream_table_dataframe_to_string(stream_table))



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Database operations API
"""

# stdlib
import logging
import re
from typing import Dict, List, Optional, Union

# third-party
try:
    import certifi
except ImportError:
    certifi = None
from pymongo import MongoClient
from pymongo.errors import ConnectionFailure

# package
from .data_model import Result, Component, Reaction, Base, DataWrapper
from .error import BadConfiguration

__author__ = "Dan Gunter (LBNL)"

_log = logging.getLogger(__name__)


class ElectrolyteDB:
    """Interface to the Electrolyte database.

    This uses MongoDB as the underlying data store.
    """

    DEFAULT_HOST = "localhost"
    DEFAULT_PORT = 27017
    DEFAULT_URL = f"mongodb://{DEFAULT_HOST}:{DEFAULT_PORT}"
    DEFAULT_DB = "electrolytedb"

    # Default timeout, in ms, for sockets, connections, and server selection
    timeout_ms = 5000
    timeout_args = {
        "socketTimeoutMS": timeout_ms,
        "connectTimeoutMS": timeout_ms,
        "serverSelectionTimeoutMS": timeout_ms,
    }

    # make sure these match lowercase names of the DataWrapper subclasses in
    # the `data_model` module
    _known_collections = ("base", "component", "reaction")

    def __init__(
        self,
        url: str = DEFAULT_URL,
        db: str = DEFAULT_DB,
        check_connection: bool = True,
    ):
        """Constructor.

        Args:
            url: MongoDB server URL
            db: MongoDB 'database' (namespace) to use
            check_connection: If True, check immediately if we can connect to the
                server at the provided url. Otherwise defer this check until the
                first operation (at which point a stack trace may occur).

        Raises:
            pymongo.errors.ConnectionFailure: if check_connection is True,
                 and the connection fails
        """
        self._mongoclient_connect_status = {"initial": "untried", "retry": "untried"}
        self._client = self._mongoclient(url, check_connection, **self.timeout_args)
        if self._client is None:
            msg = self.connect_status_str
            _log.error(msg)
            raise ConnectionFailure(msg)
        self._db = getattr(self._client, db)
        self._db = getattr(self._client, db)
        self._database_name = db
        self._server_url = url

    def is_empty(self) -> bool:
        if self._database_name not in self._client.list_database_names():
            return True
        collections = set(self._db.list_collection_names())
        if not collections:
            return True
        if not {"base", "component", "reaction"}.intersection(collections):
            _log.warning(
                "Bootstrapping into non-empty database, but without any EDB collections"
            )
            return True
        return False

    @staticmethod
    def drop_database(url, db):
        """Drop a database.

        Args:
            url: MongoDB server URL
            db: Database name

        Returns:
            None

        Raises:
            anything pymongo.MongoClient() can raise
        """
        client = MongoClient(host=url)
        client.drop_database(db)

    @classmethod
    def can_connect(cls, url=None, db=None) -> bool:
        """Convenience method to check if a connection can be made without having
           to instantiate the database object.

        Args:
            url: Same as constructor
            db: Same as constructor

        Returns:
            True, yes can connect; False: cannot connect
        """
        url = url or f"mongodb://{cls.DEFAULT_HOST}:{cls.DEFAULT_PORT}"
        db = db or cls.DEFAULT_DB
        result = True
        try:
            _ = cls(url=url, db=db, check_connection=True)
        except ConnectionFailure:
            result = False
        return result

    def _client_can_connect(self, client: MongoClient) -> bool:
        # NOTE the "ping" command is chosen because it's the only one available when using mocked MongoClient instances
        # therefore, having a single commands that works for both mock- and non-mock objects makes the mocking easier
        server_resp = client.admin.command("ping")
        try:
            return bool(server_resp["ok"])
        except (KeyError, TypeError) as e:
            _log.exception(f"Unexpected format for server response: {server_resp}")
        return None

    def _mongoclient(self, url: str, check, **client_kw) -> Union[MongoClient, None]:
        _log.debug(f"Begin: Create MongoDB client. url={url}")
        mc = MongoClient(url, **client_kw)
        if not check:
            _log.info(f"Skipping connection check for MongoDB client. url={url}")
            _log.debug(f"End: Create MongoDB client. url={url}")
            return mc
        # check that client actually works
        _log.info(f"Connection check MongoDB client url={url}")
        try:
            if self._client_can_connect(mc):
                self._mongoclient_connect_status["initial"] = "ok"
            _log.info("MongoDB connection succeeded")
        except ConnectionFailure as conn_err:
            mc = None
            self._mongoclient_connect_status["initial"] = str(conn_err)
            if "CERTIFICATE_VERIFY_FAILED" in str(conn_err):
                _log.warning(
                    f"MongoDB connection failed due to certificate " f"verification."
                )
                if certifi is not None:
                    _log.info(
                        "Retrying MongoDB connection with explicit location "
                        f"for client certificates ({certifi.where()})"
                    )
                    try:
                        mc = MongoClient(url, tlsCAFile=certifi.where(), **client_kw)
                        if self._client_can_connect(mc):
                            _log.info("Retried MongoDB connection succeeded")
                    except ConnectionFailure as err:
                        mc = None
                        self._mongoclient_connect_status["retry"] = str(err)
                        _log.error(self.connect_status_str)
        _log.debug(f"End: Create MongoDB client. url={url}")
        return mc

    @property
    def connect_status(self) -> Dict:
        return self._mongoclient_connect_status.copy()

    @property
    def connect_status_str(self) -> str:
        e = self._mongoclient_connect_status
        if e["initial"] == "ok":
            return "Connection succeeded"
        if e["retry"] == "ok":
            return "Initial connection failed, but retry succeeded"
        return f"Initial connection error ({e['initial']}), retry error ({e['retry']})"

    @property
    def database(self):
        return self._database_name

    @property
    def url(self):
        return self._server_url

    def get_components(
        self,
        component_names: Optional[List[str]] = None,
        element_names: Optional[List[str]] = None,
    ) -> Result:
        """Get thermodynamic information for components of reactions.

        Args:
            component_names: List of component names
            element_names: List of element names (ignored if component_names is given)

        Returns:
            All components matching the criteria (or all if none specified)
        """
        collection = self._db.component
        if component_names:
            query = {"$or": [{"name": n} for n in component_names]}
            _log.debug(f"get_components. components={component_names} query={query}")
            it = collection.find(filter=query)
        elif element_names:
            elt_set, elt_list = set(element_names), list(element_names)
            # Find all components with at least one of the specified elements,
            # then filter results to include only components where the elements
            # are a subset of the specified elements (i.e., no 'other' elements).
            it = (
                doc
                for doc in collection.find({"elements": {"$in": elt_list}})
                if set(doc["elements"]) <= elt_set
            )
        else:
            _log.debug(f"get_components. get all components (empty query)")
            it = collection.find(filter={})
        result = Result(iterator=it, item_class=Component)
        return result

    def get_reactions(
        self,
        component_names: Optional[List] = None,
        phases: Union[List[str], str] = None,
        any_components: bool = False,
        include_new_components: bool = False,
        reaction_names: Optional[List] = None,
    ) -> Result:
        """Get reaction information.

        Args:
            component_names: List of component names
            phases: Phase(s) to include; if not given allow any.
            any_components: If False, the default, only return reactions where
               one side of the reaction has all components provided.
               If true, return the (potentially larger) set of reactions where
               any of the components listed are present.
            include_new_components: If False, the default, only return reactions where
               all given components are found in that reaction (and no new components)
               are used in that reaction.
            reaction_names: List of reaction names instead of component names

        Returns:
            All reactions containing any of the names (or all reactions,
            if not specified)
        """
        collection = self._db.reaction
        if component_names:
            found = []
            if phases is None:
                allow_phases = None
            elif isinstance(phases, str):
                allow_phases = {phases}
            else:
                allow_phases = set(phases)
            # build a set of normalized component names
            cnames = {c.replace(" ", "_") for c in component_names}
            _log.debug(
                f"Get reaction with {'any' if any_components else 'all'} "
                f"components {cnames}"
            )
            # Brute force table scan: need to restructure DB for this to be
            # easy to do with a MongoDB query, i.e. need to put all the
            # *keys* for stoichiometry.Liq as *values* in an array, then do a:
            # {$not: {$elemMatch: { $nin: [<components>] } } } on that array
            stoich_field = Reaction.NAMES.stoich
            for item in collection.find():
                stoich = {}
                disallow = False
                for phase in item[stoich_field].keys():
                    if allow_phases is not None and phase not in allow_phases:
                        disallow = True
                    for n in item[stoich_field][phase]:
                        stoich[n] = item[stoich_field][phase][n]
                # If the item involves a phase that is not allowed, then move on to next item
                if disallow:
                    continue
                # If stoich is empty, then move on to next item
                if stoich == {}:
                    continue
                if any_components:
                    # look for non-empty intersection
                    if set(stoich.keys()) & cnames:
                        found.append(item)
                else:
                    # ok if it matches both sides
                    if set(stoich.keys()) == cnames:
                        found.append(item)
                    # also ok if it matches everything on one side
                    else:
                        # Add a reaction if all the products/reactants
                        #   can be formed. This allows addition of reactions
                        #   that may include species not yet considered.
                        if include_new_components == True:
                            for side in -1, 1:
                                side_keys = (
                                    k for k, v in stoich.items() if abs(v) / v == side
                                )
                                if set(side_keys).issubset(cnames):
                                    found.append(item)
                                    break  # found; stop
                        # Otherwise, only include reactions that are subsets of
                        #   the given component list
                        else:
                            if set(stoich.keys()).issubset(cnames):
                                found.append(item)
            it = iter(found)
        elif reaction_names:
            query = {"name": {"$in": reaction_names}}
            _log.debug(f"reaction query: {query}")
            it = collection.find(filter=query)
        else:
            it = collection.find()
        return Result(iterator=it, item_class=Reaction)

    def get_base(self, name: str = None) -> Union[Result, Base]:
        """Get base information by name of its type.

        Args:
            name: Name of the base type.
        Returns:
            If no name is given, a Result iterator over all the bases.
            Otherwise, a single `Base` object.
        """
        if name:
            query = {"name": name}
        else:
            query = {}
        collection = self._db.base
        result = Result(iterator=collection.find(filter=query), item_class=Base)
        if name:
            try:
                return list(result)[0]
            except IndexError:
                raise IndexError("No bases found in DB")
        else:
            return result

    def list_bases(self):
        """List the currently loaded bases and provide brief description

        Args:
            None
        Returns:
            No return, just display info to console
        """
        for item in self.get_base():
            print(
                f"base name: {item.name}\t\tdescription -> {self._base_desc(item.name)}"
            )

    def _base_desc(self, name) -> str:
        """Creates a description of a base based on the standard naming

        Args:
            name: Name of the base to describe
        Returns:
            desc: String of the description of the base
        """
        if name == "default_thermo":
            desc = "ThermoConfig: Default uses FTPx state vars for Liq phase"
        elif name == "reaction":
            desc = "ReactionConfig: Blank reaction template"
        else:
            items = name.split("_")
            if len(items) < 3:
                raise BadConfiguration(
                    "ElectrolyteDB._base_desc",
                    self.get_base(name).idaes_config,
                    missing=None,
                    why="\nName of base (" + name + ") is of unknown format\n",
                )
            if items[0] == "thermo":
                desc = "ThermoConfig: "
            else:
                desc = "ReactionConfig: "
            desc += "uses " + items[-1] + " state vars for "
            for i in range(1, len(items) - 1):
                desc += items[i] + ","
            desc += " phases"

        return desc

    # older method name
    get_one_base = get_base

    def load(
        self,
        data: Union[Dict, List[Dict], DataWrapper, List[DataWrapper]],
        rec_type: str = "base",
    ) -> int:
        """Load a single record or list of records.

        Args:
            data: Data to load, as a single or list of dictionaries or :class:`DataWrapper` subclass
            rec_type: If input is a dict, the type of record. This argument is ignored if the input is
                      a subclass of DataWrapper.

        Returns:
            Number of records loaded
        """
        is_object = False
        if isinstance(data, DataWrapper):
            data = [data]
            is_object = True
        elif isinstance(data, dict):
            data = [data]
        else:
            is_object = isinstance(data[0], DataWrapper)
        if is_object:
            rec_type = data[0].__class__.__name__.lower()
        else:
            assert rec_type in self._known_collections
        num = 0
        for item in data:
            coll = getattr(self._db, rec_type)
            record = item.json_data if is_object else item
            coll.insert_one(self.preprocess_record(record, rec_type))
            num += 1
        return num

    # TODO: This preprocessing overlaps with data_model.DataWrapper subclasses.
    #       It should all be moved to one place. Unclear how to accomplish this
    #       without also disrupting how structure is validated. The _preprocess
    #       functions in each data_model.DataWrapper do not explicitly perform
    #       these actions and do not return anything. Some of the data_model.
    #       DataWrapper objects do not have a _preprocess function.

    # record is a dict and rec_type is a string
    @classmethod
    def preprocess_record(cls, record, rec_type):
        process_func = getattr(cls, f"_process_{rec_type}")
        return process_func(record)

    # You each of these are needed because they get checked in 'validate.py'
    @staticmethod
    def _process_component(rec):
        # This line is needed because it is checked for in structure
        rec["elements"] = get_elements_from_components([rec["name"]])
        return rec

    @staticmethod
    def _process_reaction(rec):
        # These lines seem integral to loading the database
        # ------------------------------------------------
        # If reaction_order is not present in parameters, create it by
        # copying the stoichiometry (or empty for each phase, if stoich. not found)
        if Reaction.NAMES.param in rec:
            param = rec[Reaction.NAMES.param]
            if Reaction.NAMES.reaction_order not in param:
                if Reaction.NAMES.stoich in rec:
                    param[Reaction.NAMES.reaction_order] = rec[
                        Reaction.NAMES.stoich
                    ].copy()
                else:
                    param[Reaction.NAMES.reaction_order] = {
                        phase: {} for phase in Reaction.PHASES
                    }

        return rec

    # This function does nothing... but gets called
    @staticmethod
    def _process_base(rec):
        return rec

    # This function appears to be done implicitly by data_model.Component
    # and is never actually called from the above 'load' function
    @staticmethod
    def _process_species(s):
        """Make species match https://jess.murdoch.edu.au/jess_spcdoc.shtml"""
        m = re.match(r"([a-zA-Z0-9]+)\s*(\d*[\-+])?", s)
        if m is None:
            raise ValueError(f"Bad species: {s}")
        symbols, input_charge = m.groups()
        if input_charge is None:
            charge = ""
        elif len(input_charge) > 1:
            # make 2+ -> +2
            num = input_charge[:-1]
            sign = input_charge[-1]
            charge = f"{sign}{num}"
        else:
            charge = input_charge
        return f"{symbols}{charge}"


# Helper function used by _process_component above
def get_elements_from_components(components):
    elements = set()
    for comp in components:
        for m in re.finditer(r"[A-Z][a-z]?", comp):
            element = comp[m.start() : m.end()]
            if element[0] == "K" and len(element) > 1:
                pass
            else:
                elements.add(element)
    return list(elements)
